/*
Copyright (c) 2007 Gerard Marull Paretas.
*/

#include <avr/io.h>

#include "onewire.h"

/*
	Note: This library seams to work only with -O3 compiler optimization flag.
*/

inline __attribute__((gnu_inline)) void onewire_delay(uint16_t delay)
{
	while(delay--) asm volatile("nop");
}

uint8_t onewire_reset()
{
	uint8_t i;

	//Pull line low and wait for 480uS
	THERM_LOW();
	THERM_OUTPUT_MODE();
	onewire_delay(us(480));

	//Release line and wait for 60uS
	THERM_INPUT_MODE();
	onewire_delay(us(60));

	//Store line value and wait until the completion of 480uS period
	i = (THERM_PIN & (1<<THERM_DQ));
	onewire_delay(us(420));

	//Return the value read from the presence pulse (0=OK, 1=WRONG)
	return i;
}

void onewire_wbit(uint8_t bit)
{
	//Pull line low for 1uS
	THERM_LOW();
	THERM_OUTPUT_MODE();
	onewire_delay(us(1));

	//If we want to write 1, release the line (if not will keep low)
	if(bit) THERM_INPUT_MODE();

	//Wait for 60uS and release the line
	onewire_delay(us(60));
	THERM_INPUT_MODE();
}

uint8_t onewire_rbit(void)
{
	uint8_t bit=0;

	//Pull line low for 1uS
	THERM_LOW();
	THERM_OUTPUT_MODE();
	onewire_delay(us(1));

	//Release line and wait for 14uS
	THERM_INPUT_MODE();
	onewire_delay(us(14));

	//Read line value
	if(THERM_PIN&(1<<THERM_DQ)) bit=1;

	//Wait for 45uS to end and return read value
	onewire_delay(us(45));
	return bit;
}

uint8_t onewire_rbyte(void)
{
	uint8_t i=8, n=0;
	while(i--)
	{
		//Shift one position right and store read value
		n>>=1;
		n|=(onewire_rbit()<<7);
	}
	return n;
}

void onewire_wbyte(uint8_t byte)
{
	uint8_t i=8;
	while(i--)
	{
		//Write actual bit and shift one position right to make the next bit ready
		onewire_wbit(byte&1);
		byte>>=1;
	}
}

int8_t therm_read_temperature()
{
	int16_t rez;

	// reset
	if(onewire_reset() != 0)
		return THERM_INVALID; // return the "invalid" value

	// double-reset, from possible previous not successful reset
	if(onewire_reset() != 0)
		return THERM_INVALID; // return the "invalid" value

	onewire_wbyte(THERM_CMD_SKIPROM); // skip ROM
	onewire_wbyte(THERM_CMD_CONVERTTEMP); // start temperature conversion
	while(!onewire_rbit()); // wait until conversion is complete, as per DS18S20 datasheet

	// reset
	if(onewire_reset() != 0)
		return THERM_INVALID; // return the "invalid" value
	onewire_wbyte(THERM_CMD_SKIPROM); // skip ROM
	onewire_wbyte(THERM_CMD_RSCRATCHPAD); // send command to read Scratchpad

	// read bytes from scratchpad (only 2 first bytes)
	uint8_t ls_byte = onewire_rbyte(); // get the temperature in 0.5 degree steps
	rez = onewire_rbyte(); // get the "sign" byte into the lower resulting byte for now
	rez = rez << 8; // move it up!
	rez |= ls_byte; // now add the lower byte and that's it

	// reset to cancel further reading of Scratchpad memory
	onewire_reset(); // don't test the return value - it is not important

	// divide by 2 to get the real temperature in degrees, not in 0.5 steps
	//rez = rez / 2; // this we will do outside the function

	return (int8_t)rez; // it can all fit into 8 bits
}
