/***************************************************
 *
 * "THE DON'T-CARE LICENCE"
 * I don't care, do whatever you want with this file.
 *
 * www.elektronika.ba
 *
****************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdlib.h> // just for random nrs in "Candle Mode"
#include <avr/eeprom.h>

#include "wt1.h"
#include "onewire.h" // <- Copyright (c) 2007 Gerard Marull Paretas.

// HV booster functions
void boost_init(void); // sets uC pins and HW
uint8_t boost_go(uint8_t, uint8_t); // starts the booster, returns 0 on success
void boost_halt(void); // stops the booster
uint8_t boost_feedback(uint8_t); // gets booster output voltage on call
void boost_restart(void); // this encapsulates booster functions and also reports result on LEDs and Nixies
void boost_continue(void); // re-enables the booster PWM and calls boost_restart()

// Sound generation functions
void sound_init(void); // sets uC pins and HW
void sound_beep(uint8_t, uint16_t); // beeps @freq for dur duration (ms)

// Sleep start/end functions
void sleep_go(void); // sets all to sleep mode, turn off some peripherals
void sleep_end(void); // restores all from sleep mode, turn back on some peripherals

// RTC functions
void rtc_init(void); // inits the timers for RTC (Timer/Counter2)
uint8_t isleapyear(uint16_t); // calculate if given year is leap year
void alarm_on(void); // turning alarm on
void alarm_off(void); // turning alarm off
void alarm_snooze(void); // snoozing the alarm
void alarm_play(void); // set melody to play

// Pin interrupts setup
void pin_interrupt_init(void); // init IO pins used with interrupts and set them

// Interrupt driven delay 1-~65536ms (~65sec max)
void delay_ms_(uint16_t); // interrupt based delay function

// TLC59401 related functions
void tlc59401_init(void); // sets uC pins and HW
void tcl59401_ClockInDC(void); // sending DC data to TLC chip
void tcl59401_ClockInGS(void); // sending GS data to TLC chip
void tlc59401_SetGSdata(uint8_t, uint16_t); // sets GS data for a channel in data array
void rgb_led(uint8_t, uint8_t, uint8_t, uint8_t); // sets "nr" RGB LED to 16* r,g,b value; nr is 0..3

// SN74141 related functions
void sn74141_init(void); // sets uC pins and HW
void sn74141_set(uint8_t); // sets one of 10 outputs based on numerical input; digit is 0..9

// Nixie tube related functions
void set_brightness_of(uint8_t, uint16_t); // if tube_idx >= DISPLAY_SIZE then brightness is set to all available tubes

// Misc. functions
void display_time(void); // fills-in display_digits[]s with HH:MM
void display_temp(uint16_t); // reads DS18S20's temperature and fills-in display_digits[]s with DD'D and holds there for duration parameter
void display_date(void); // displays full date on 4 digits
void display_menu(void); // display menu for setup
void set_rtc(void); // setting the time and date
void set_alarm_time(void); // setting the time of alarm
void set_mode_times(void); // setting mode durations
void set_brightness(void); // set display brightness
void set_snooze_interval(void); // set snooze interval in minutes
void set_time_led_prefs(void); // set bottom leds toggling and their color for TIME
void set_temp_blender(void); // set temperature color blender on/off
void set_date_led_prefs(void); // set bottom leds on/off and their color for DATE
void set_date_fading(void); // set tubes fading value and on/off for DATE display
void set_auto_off_prefs(void); // set auto-off times and on/off
void set_cross_fader_prefs(void); // set cross-fader speed and on/off
void restore_setup(void); // activate user preferences to application

void faded_chase_leds(uint8_t, uint8_t, uint8_t, uint8_t, uint8_t, uint8_t, uint8_t, uint16_t, uint8_t); // chasing LEDs with fade effect with a custom color (NOTE: delay is not very acurate)
uint8_t faded_chase_leds_calc(uint8_t, uint8_t, uint8_t, uint8_t, uint8_t); // a "helper" function of faded_chase_leds() function
void candle_mode(void); // special mode for a bit of romance
uint8_t rgb_blender(uint8_t, uint8_t, int8_t); // makes a real color for a given temperature
void disp_next(void); // moving to next thing to display

// Globals
volatile uint8_t timer0a_step = 0; // counting from 0...15 and is reset in ISR
volatile uint8_t timer0b_step = 0; // counting from 0...31 and is reset in ISR
volatile uint8_t display_refresh = 0; // display refresh rate counter for the ISR
volatile uint16_t milliseconds = 0; // uC millisecond counter used in delay_ms_(), and is reset throughout the application
volatile uint16_t max_brightness = DISP_MAX_BRIGHTNESS; // top brightness at the moment - this is managed by user settings
volatile uint16_t brightness[DISPLAY_SIZE] = { DISP_MAX_BRIGHTNESS, DISP_MAX_BRIGHTNESS, DISP_MAX_BRIGHTNESS, DISP_MAX_BRIGHTNESS }; // brightness for each tube and is maximum at start for all
volatile uint8_t display_digits[DISPLAY_SIZE]; // digits for the tubes in HH:MM order
volatile uint8_t disp_digit_idx = 0; // which digit is shown in multiplex at the time [0..3 for 4 tubes] of DISPLAY_SIZE
volatile uint8_t display_digit; // current digit that is displayed (sent) to the SN74141

#ifdef CROSS_FADING_CODE
volatile uint8_t display_digits_sys[DISPLAY_SIZE]; // digits for the tubes in HH:MM order for the actually showing - used only in ISR
volatile uint16_t cross_fade_brightness[DISPLAY_SIZE] = { DISP_LAST_OFF_BRTNS+1, DISP_LAST_OFF_BRTNS+1, DISP_LAST_OFF_BRTNS+1, DISP_LAST_OFF_BRTNS +1}; // brightness for cross-fading.. initialized with +1 just for firs use - IMPORTANT
volatile uint8_t cross_fade_tmp[DISPLAY_SIZE]; // introduced this array just to speed up the cross-fading effect
volatile uint8_t cross_fade_speed = DEFAULT_CROSS_FAD_SPD; // cross fading speed set by user
#endif

volatile uint8_t disp_duration_timer = 0; // timer for displaying stuff
volatile uint8_t disp_duration[DISP_THINGS] = { DEFAULT_DISP_0, DEFAULT_DISP_1, DEFAULT_DISP_2 }; // defaults .. in seconds
volatile uint8_t disp_what = 0; // what to display now

volatile uint16_t snooze_timer = 0; // for snoozing in seconds
volatile uint8_t snooze_auto_executer = 0; // auto executer of snooze timer after it reaches RTC_AUTO_SNOOZE value
volatile uint8_t snooze_interval = DEFAULT_SNOOZE_INTERVAL; // duration of snooze in minutes

volatile uint8_t rgb_time_color[3] = { TimeLEDColorR, TimeLEDColorG, TimeLEDColorB }; // defaults
volatile uint8_t rgb_date_color[3] = { DateLEDColorR, DateLEDColorG, DateLEDColorB }; // defaults

volatile uint16_t date_fading_brightness = DISP_MIN_BRIGHTNESS;	// default

volatile uint16_t auto_off_brightness = DISP_MAX_BRIGHTNESS; // when auto-off executes, this goes down to 0. when wakes up, it goes back to max_brightness
volatile uint8_t auto_off_prescaler; // to slow down the fade-out
volatile uint8_t auto_off_auto_executer = 0; // if >0 then this is the time in seconds when auto-off will be executed

// sound generation related stuff
volatile uint8_t sound_presc = 0;
volatile uint16_t sound_duration = 0;
volatile uint8_t sound_freq = 0;
volatile uint8_t sound_melody_idx = 0;

// button related globals:
volatile uint8_t prev_btn = 0; // previous button states for interrupt handling
volatile uint8_t btn_press = 0; // also for the buttons with the isr
volatile uint8_t btn_hold = 0; // also for the buttons with the isr
volatile uint16_t btn_hold_timer = 0; // to detect button holds
volatile uint16_t btn_inactivity_timer = 0; // to detect button timeouts

volatile uint8_t bools[BOOL_BANK_SIZE] = { 0, 0, 0, 0 }; // 8 global boolean values per BOOL_BANK_SIZE

// DC - dot correction data for the TLC59401 chip - To save SRAM, this can go to PROGMEM since we don't update it rapidly like tlc_GSdata
volatile uint8_t tlc_DCdata[12] = {
	255,255,255,255,255,255,255,255,255,255,255,255
};
// GS - grayscale data for the TLC59401 chip
volatile uint8_t tlc_GSdata[24] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};
volatile uint8_t RTC[6] = { 12, 0, 0, 1, 1, 10 }; // we have: H,M,S,d,m,y
volatile uint8_t RTC_ALARM[2] = { 0, 0 }; // we have: H,M
volatile uint8_t RTC_AOFF[2] = { 0, 0 }; // clock automatic OFF, we have: H,M
volatile uint8_t RTC_AON[2] = { 0, 0 }; // clock automatic ON, we have: H,M

//############
// Main app //
//############
int main(void)
{
	// turn on pin-change-interrupts
	pin_interrupt_init(); // this will set the DDRs for buttons, switch and pwr-sense pin and prepare the interrupts

	// if there is no power at the start (if the battery is first power for us) wait untill we get some!
	while( !(PWR_SENS_PINREG & _BV(PWR_SENS_PIN)) );

	// SN74141 init
	sn74141_init();

	// TLC init
	tlc59401_init(); // init TLC chip hw and Timer1 Comp-A for 4096 blanking signal
	tcl59401_ClockInDC(); // init the DCs. GS will be sent by first display refresh interval from interrupt

	// Interrupts ON
	sei();

	// HV Booster init & execute
	boost_init();										// init the booster
	boost_restart();									// starts the booster, and reports voltage on Nixies and LEDs

	// RTC init
	rtc_init(); 										// Cesium counter on :)

	// Check to see if there is something valid in EEPROM
    uint8_t *pEE = (uint8_t *)EEPROM_START;
	if( eeprom_read_byte(pEE++) == 0xAA ) // 0xAA is my "valid flag"
	{
		eeprom_read_block((void *)&RTC,pEE,6);
		pEE+=6;
		eeprom_read_block((void *)&RTC_ALARM,pEE,2);
		pEE+=2;
		eeprom_read_block((void *)&RTC_AOFF,pEE,2);
		pEE+=2;
		eeprom_read_block((void *)&RTC_AON,pEE,2);
		pEE+=2;
		eeprom_read_block((void *)&rgb_time_color,pEE,3);
		pEE+=3;
		eeprom_read_block((void *)&rgb_date_color,pEE,3);
		pEE+=3;
		max_brightness = eeprom_read_word((uint16_t *)pEE);
		pEE+=2;
		date_fading_brightness = eeprom_read_word((uint16_t *)pEE);
		pEE+=2;
		snooze_interval = eeprom_read_byte(pEE++);
		eeprom_read_block((void *)&disp_duration,pEE,DISP_THINGS);
		pEE+=DISP_THINGS;
		/*
		eeprom_read_block((void *)&bools,pEE,BOOL_BANK_SIZE);
		bools[BSYS]=0x00; // clear this !
		bools[BAPP1]=0x00; // clear this !
		bools[BAPP2]=0x00; // clear this !
		*/
		bools[BUSR1] = eeprom_read_byte(pEE);
	}
	else
	{
		// Set "clock release" time and date
		RTC[TIME_H]=11;
		RTC[TIME_M]=00;
		RTC[TIME_S]=00;
		RTC[DATE_D]=27;
		RTC[DATE_M]=11;
		RTC[DATE_Y]=10;

		bs(RTC_IsSet,BAPP1);
		bc(BlankTime,BAPP1);
		bs(LEDtoggling,BUSR1);
		// --
	}

	// Activate user preferences to application
	restore_setup();

	// Init the sounder
	sound_init();

	// hello sound
	sound_beep(150,25);
	sound_beep(70,30);
	sound_beep(20,40);
	sound_beep(5,10);

	// hello leds
	faded_chase_leds(0,0,0, 20,0,250, 10, 1000, 2);	// lets try faded chasing BLUISH for fun, backwards

	// Main app loop
	while(1)
	{
		// here we check if we entered the sleep mode. if so, we don't do anything until we get some power!
		while( bv(SleepMode,BSYS) );					// stay here until power arrives

		// Button press checker in "idle" mode
		switch(btn_press | btn_hold)
		{
			// BTN-A was pressed OR held: DATE or MENU
			case 0b00000001:
				if(btn_press & 0b00000001)			// was it a press?
				{
					btn_press &= ~0b00000001; 		// button handled
					display_date(); 					// show date
				}
				else									// nope, it was a hold
				{
					btn_hold &= ~0b00000001; 			// button handled
					display_menu(); 					// show menu
				}
			break;

			// BTN-B was pressed OR held: TEMPERATURE or CANDLE MODE
			case 0b00000010:
				if(btn_press & 0b00000010)			// was it a press?
				{
					btn_press &= ~0b00000010; 		// button handled
					display_temp(3000); 				// show date
					restore_setup();					// activate user preferences to application
					set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
					set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);
					rgb_led(0,0,0,0);
					rgb_led(1,0,0,0);
					rgb_led(2,0,0,0);
					rgb_led(3,0,0,0);
				}
				else									// nope, it was a hold
				{
					btn_hold &= ~0b00000010; 			// button handled
					candle_mode(); 						// candle mode
				}
			break;
		}

		if( bv(AutoOffed,BSYS) ) continue;				// don't go below if auto-offed

		// Displaying loop
		switch(disp_what)
		{
			case 0:										// TIME
				display_time();

				if(disp_duration_timer > disp_duration[disp_what]) disp_next(); // switch to next thing to display ?
			break;

			case 1:										// DATE
				display_date();

				if(disp_duration_timer > disp_duration[disp_what]) disp_next(); // switch to next thing to display ?
			break;

			case 2:										// TEMPERATURE
				display_temp(333);						// show and slow down

				if(disp_duration_timer > disp_duration[disp_what])
				{
					disp_next(); // switch to next thing to display ?
					set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
					set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);
					restore_setup();					// activate user preferences to application
					rgb_led(0,0,0,0);
					rgb_led(1,0,0,0);
					rgb_led(2,0,0,0);
					rgb_led(3,0,0,0);
				}
			break;
		}

	}

	return 0;
}


//##############################
// Interrupt: TIMER2 OVERFLOW //
//##############################
// called once every second
ISR(TIMER2_OVF_vect, ISR_NOBLOCK) // this interrupt routine can be interrupted by any other, therefore this interrupt routine must complete in under a second !!!
{
	/*
	To avoid unintentional changes of clock frequency, a special write procedure must be followed
	to change the CLKPS bits:
		1. Write the Clock Prescaler Change Enable (CLKPCE) bit to one and all other bits in
		CLKPR to zero.
		2. Within four cycles, write the desired value to CLKPS while writing a zero to CLKPCE.
		Interrupts must be disabled when changing prescaler setting to make sure the write procedure is
		not interrupted.
	*/
	//CLKPR = _BV(CLKPCE); - OVO CE MI TREBATI KADA BUDEM PRAVIO SLEEP() KOJI CE SMANJITI CLOCK FREQ. ZA MANJU POTROSNJU
	//CLKPR = 0;

	RTC[TIME_S]++; // one second has gone by
	//RTC[TIME_S]+=5; // speedy clock for testing things

	// a minute!
	if(RTC[TIME_S] >= 60)
	{
		RTC[TIME_S] = 0;
		RTC[TIME_M]++;
	}

	// an hour...
	if(RTC[TIME_M] >= 60)
	{
		RTC[TIME_M] = 0;
		RTC[TIME_H]++;
	}

	// a day....
	if(RTC[TIME_H] >= 24)
	{
		RTC[TIME_H] = 0;
		RTC[DATE_D]++;
	}

	// a full month with leap year checking!
	if(
		(RTC[DATE_D] > 31)
		|| (
			(RTC[DATE_D] == 31)
			&& (
				(RTC[DATE_M] == 4)
				|| (RTC[DATE_M] == 6)
				|| (RTC[DATE_M] == 9)
				|| (RTC[DATE_M] == 11)
				)
		)
		|| (
			(RTC[DATE_D] == 30)
			&& (RTC[DATE_M] == 2)
		)
		|| (
			(RTC[DATE_D] == 29)
			&& (RTC[DATE_M] == 2)
			&& !isleapyear(2000+RTC[DATE_Y])
		)
	)
	{
		RTC[DATE_D] = 1;
		RTC[DATE_M]++;
	}

	// HAPPY NEW YEAR!
	if(RTC[DATE_M] >= 13)
	{
		RTC[DATE_Y]++;
		RTC[DATE_M] = 1;
	}

	// if we are sleeping, don't do anything below... no use of it :)
	if( bv(SleepMode,BSYS) )
	{
		return;
	}

	// Syncrhonized toggling of dots and LEDs
	if( RTC[TIME_S] % 2 && !bv(AutoOffed,BSYS) ) // dont work LEDs while auto-offed
	{
		if( bv(ToggleLEDs,BAPP1) )
		{
			rgb_led(0,rgb_time_color[0],rgb_time_color[1],rgb_time_color[2]);
			rgb_led(1,rgb_time_color[0],rgb_time_color[1],rgb_time_color[2]);
			rgb_led(2,rgb_time_color[0],rgb_time_color[1],rgb_time_color[2]);
			rgb_led(3,rgb_time_color[0],rgb_time_color[1],rgb_time_color[2]);
		}
		if( bv(ToggleColons,BAPP1) )
		{
			set1(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
			set1(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);
		}
		if( !bv(RTC_IsSet,BAPP1) )
		{
			bs(BlankTime,BAPP1);
		}
	}
	else
	{
		if( bv(ToggleLEDs,BAPP1) )
		{
			rgb_led(0,0,0,0);
			rgb_led(1,0,0,0);
			rgb_led(2,0,0,0);
			rgb_led(3,0,0,0);
		}
		if( bv(ToggleColons,BAPP1) )
		{
			// keep the upper dot ON while snooze is active!
			if( !bv(RTC_Snoozing,BAPP1) )
			{
				set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);	// snooze not active, so turn this one off also so it actually blinks
			}
			set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);		// the bottom one will be blinking...
		}
		if( !bv(RTC_IsSet,BAPP1) )
		{
			bc(BlankTime,BAPP1);
		}
	}

	// time to go to Auto Off? - shutdown until regular wakeup or alarm?
	if( bv(AutoOffMode,BUSR1)
		&& (RTC_AOFF[TIME_H] == RTC[TIME_H])
		&& (RTC_AOFF[TIME_M] == RTC[TIME_M])
		&& (RTC[TIME_S] == 0) )
	{
		bs(AutoOffed,BSYS);

		// added 02-11-2010
		//boost_halt(); // halt the booster
		//--
	}
	// time to go to Auto off by automatic executer?
	if(auto_off_auto_executer)
	{
		auto_off_auto_executer--;
		if(auto_off_auto_executer==0)
		{
			bs(AutoOffed,BSYS);

			//boost_halt(); // halt the booster
		}
	}
	// time to go out of Auto Off? - wakeup the tubes!
	if( bv(AutoOffed,BSYS)
		&& (RTC_AON[TIME_H] == RTC[TIME_H])
		&& (RTC_AON[TIME_M] == RTC[TIME_M])
		&& (RTC[TIME_S] == 0) )
	{
		bc(AutoOffed,BSYS);

		// added 02-11-2010
		auto_off_auto_executer=0;
		boost_continue(); // continue the booster
		//--
	}
	// added 02-11-2010
	// if auto-off did its fade-out-countdown, shut down the booster also
	if( bv(AutoOffed,BSYS) && !auto_off_brightness)
	{
		boost_halt(); // halt the booster
	}
	//--

	// set the AntiPoison boolean if it is time to anti-poison the nixies, on every 15 minutes
	if( RTC[TIME_S] == 0 && (RTC[TIME_M] == 0 || RTC[TIME_M] == 15 || RTC[TIME_M] == 30 || RTC[TIME_M] == 45) )
	{
		bs(AntiPoison,BAPP1);
	}

	disp_duration_timer++; 								// we need this counter for main()

	if(snooze_timer) snooze_timer--;					// countdown, but don't overflow

	// time to ring by TIME?
	if( bv(RTC_AlarmOn,BAPP1)
		&& (RTC_ALARM[TIME_H] == RTC[TIME_H])
		&& (RTC_ALARM[TIME_M] == RTC[TIME_M])
		&& (RTC[TIME_S] == 0) )
	{
		alarm_play();									// tell ISR to play the melody
	}
	// time to ring by SNOOZE end?
	if( bv(RTC_Snoozing,BAPP1) && bv(RTC_AlarmOn,BAPP1) && snooze_timer==0 )
	{
		alarm_play();									// tell ISR to play the melody
		bc(RTC_Snoozing,BAPP1);							// yo!
	}

	// How long have we been alarming? Not too long I hope...
	if( bv(RTC_Alarming,BAPP1) )
	{
		snooze_auto_executer++;
		if(snooze_auto_executer >= RTC_AUTO_SNOOZE)		// time to execute the snooze button automatically?
		{
			snooze_auto_executer=0;
			alarm_snooze();
		}
	}
}


//###########################
// Interrupt: TIMER1 OCR1A //
//###########################
// called every 4096 instruction cycles. Note: Must complete in next <4096 instructions and should not be interrupted !!!
ISR(TIMER1_COMPA_vect, ISR_BLOCK) // this interrupt routine can NOT be interrupted by any other, it is critical!
{
	// here we reset the TCL59401's internal pointer to 0
	set1(BLANK_PORT,BLANK_PIN);

	// if me must latch-in new data (DC or GS)
	if( bv(TLC_HasGSUpdate,BSYS) || bv(TLC_HasDCUpdate,BSYS) )
	{
		// whatever we clocked in TLC before, becomes active now (actually on set0(BLANK) below)
		set1(XLAT_PORT, XLAT_PIN);
		set0(XLAT_PORT, XLAT_PIN);

		// make one extra clock signal because we updated dot-correction data and if this is the GS we just latched (as per datasheed)
		if( bv(TLC_NeedsExtraCLK,BSYS) && bv(TLC_HasGSUpdate,BSYS) )
		{
			set1(SCLK_PORT, SCLK_PIN);
			set0(SCLK_PORT, SCLK_PIN);

			bc(TLC_NeedsExtraCLK,BSYS);
		}
		bc(TLC_HasDCUpdate,BSYS); 						// say that we finished the update
		bc(TLC_HasGSUpdate,BSYS); 						// say that we finished the update
	}

	set0(BLANK_PORT,BLANK_PIN);

	// because TLC is slower than SN74141, we will update the SN74141 here, right after the TLC update. this way we don't see (kind of) the "ghosting" effect in tubes
	if( bv(SN74141_Update,BSYS) )
	{
		bc(SN74141_Update,BSYS);
		sn74141_set( display_digit );
	}
}


//##############################
// Interrupt: TIMER0 OVERFLOW //
//##############################
// called @ (F_CPU/256) = 31.25 khz for 8 MHz CPU clock, init from boost_init()
ISR(TIMER0_OVF_vect, ISR_NOBLOCK) // this interrupt routine can be interrupted by any other
{
	// if we are not sleeping refresh the nixie display
	//if( !bv(SleepMode,BSYS) )
	{
		// half of millisecond (0.48ms)
		if(++timer0a_step >= 15)
		{
			timer0a_step = 0; 								// reset so we can count another 15 (~0.5ms)
			display_refresh++; 								// display refresh timer increment
		}

		// check if it is time to refresh the display?
		if(display_refresh >= DISPLAY_REFRESH_05MS) 		// this code executes every ~DISPLAY_REFRESH_05MS [1/2 ms] !
		{
			display_refresh = 0; 							// so we can count another DISPLAY_REFRESH_05MS [1/2 ms]

			// "scan" all available tubes:
			uint8_t i = 0;
			for(i=0; i<DISPLAY_SIZE; i++)
			{
				uint8_t this_brightness;

				// the nixie that will be lit-up at this time
				if(i == disp_digit_idx)
				{
					// here we have this IF, because of "Auto-Off" function
					if( bv(AutoOffed,BSYS) )
					{
						this_brightness = auto_off_brightness; // this gets decremented to 0 or restored to maximum in ISR below
					}
					else
					{
						this_brightness = brightness[i]; 		// this one is ON - originally this line is enough for display to work when Auto-Off function is not used
					}

					#ifdef CROSS_FADING_CODE
						// CROSS-FADING CODE:
						// speed of the cross-fade is determined by "DISP_LAST_OFF_BRTNS" and "timer0a_step" (DISPLAY_REFRESH_05MS) which can't be re-adjusted just to alter the cross-fading speed, so "cross_fade_speed" is introduced
						// cross-fading enabled AND digit-changed / transition didn't complete yet?
						if( bv(CrossFaderEn,BAPP2) && !bv(AutoOffed,BSYS) && (display_digits_sys[i] != display_digits[i]) )
						{
							// To speed up the fading process, we must do things other than simple "--" decrement:
							if( ((uint32_t)cross_fade_brightness[i]-cross_fade_speed) >= DISP_LAST_OFF_BRTNS)// dont decrement if we will underflow
							{
								cross_fade_brightness[i]-=cross_fade_speed;	// current tube's fading counter (fader)
							}
							else
							{
								cross_fade_brightness[i] = DISP_LAST_OFF_BRTNS;// go to allowed 0
							}
							// ended? (time to mark the end of transition for this tube)
							if( cross_fade_brightness[i] <= DISP_LAST_OFF_BRTNS )
							{
								display_digits_sys[i] = display_digits[i]; // mark the end
								cross_fade_brightness[i] = max_brightness; // initialize for second use
							}
							// on every other "display" toggle between old and new value to be shown
							cross_fade_tmp[i]++;							// just to indicate the "every other" for current tube
							if(cross_fade_tmp[i] % 2)
							{
								display_digit = display_digits_sys[i];		// "show" the old number (or the new one if we just ended)
								this_brightness = cross_fade_brightness[i]; // this one gets decreased to the "zero-show" value so we just copy it here
							}
							else
							{
								display_digit = display_digits[i]; 			// "show" the new number
								this_brightness = max_brightness-cross_fade_brightness[i]; // TODO: -1 ili +1 ce trebati. this one gets incremented from the "zero-show" value so we have to complement it to the "max_brightness" value
							}
						}
						// cross-fading not active for the current nixie tube "i"
						else
						{
							display_digit = display_digits[i]; 				// "take" the number that will be displayed on nixie "i"
						}
					#else
						// NON-CROSS-FADING CODE:
						// without cross fading code this is enough:
						display_digit = display_digits[i]; 					// "take" the number that will be displayed on nixie "i"
					#endif

					// anti-ghosting addon. we don't use cathode-blanking with the SN74141 but instead we use anode-blanking. this helps with IN-8 tubes, but with IN-16 tubes there is no need for this addon.
					// from: http://neonixie-l.grouply.com/message/46510
					/*
						1) You should have inter-digit blanking intervals. ~200uS ballpark. - THIS IS NOT DONE.
						2) You should NOT use the cathode driver for the blanking. ie sending a 'f' to
						the 74141, so that all cathodes are OFF. One cathode should always be ON. - DONE .. not anymore
						3) Use the anode drive for your blanking. During blanking, all anodes are OFF. - DONE
					*/
					// If the digit to be displayed is blank one, we are actually going to blank the ANODE, not the CATHODE.
					if(display_digit >= nixie_blank)
					{
						// I get better results when blanking both, the anode and the cathode, so here goes:
						//display_digit = 0; 				// put something valid into the digit - one cathod will always be ON
						//if(i>0) display_digit = display_digits[i-1]; // trick that needs to be done if the above line is != nixie_blank - NOT to be used while cross-fading
						this_brightness = 0; 				// turn OFF the anode instead of a cathode ! - during blanking, turn the anode OFF
					}
				}
				// others are off
				else
				{
					this_brightness = 0; 								// the rest are OFF, and they have no value :)
				}

				// address a nixie, but backwards
				tlc59401_SetGSdata(TLC_LAST_NIXIE_CH-i, this_brightness); 	// nixie "i" is on, others are off
			}

			// refresh the TLC chip with new data from above _SetGS() executions and from elsewhere in the application
			tcl59401_ClockInGS();
			bs(SN74141_Update,BSYS); 									// tell the TIMER1_COMPA_vect ISR that it needs to update the SN74141 also

			if(++disp_digit_idx >= DISPLAY_SIZE) disp_digit_idx = 0; 	// next nixie tube in row please, and check for end to start again
		}
	}

	// almost a millisecond (0.992ms)
	if(++timer0b_step >= 31)
	{
		timer0b_step = 0; 												// reset so we can count another 31 (~1ms)
		milliseconds++; 												// another ms counter, can be reset in the program - used in delay_ms_()

		// Auto-Off function executed/active?
		if( bv(AutoOffed,BSYS) )										// fading out?
		{
			if(auto_off_brightness > DISP_LAST_OFF_BRTNS)
			{
				// slow it down
				if(++auto_off_prescaler >= 15) // was 10; 02-11-2010
				{
					auto_off_prescaler = 0;
					auto_off_brightness--;								// go down to total 0 ... not anymore, now decrement to DISP_LAST_OFF_BRTNS!
					if(auto_off_brightness == DISP_LAST_OFF_BRTNS) 		// if we came to end and DISP_LAST_OFF_BRTNS>0 than set auto_off_brightness=0 !!!
					{
						auto_off_brightness = 0;
					}
				}
			}
		}
		else											// no, fading in! ("fading in" is actually not fading)
		{
			auto_off_brightness = max_brightness;		// (keep restored) restore for next fading
		}

		if(btn_inactivity_timer) btn_inactivity_timer--;// to detect button input timeouts, we shouldn't overflow!
		if(btn_hold_timer) btn_hold_timer--;			// to detect button holds, we shouldn't overflow!

		// TODO: code below should be optimized more for speed ! It MUST complete in less than 1/31250 = (0.000032) sec.
		if(sound_duration)								// let's check if the duration is about to expire so we can turn the output off or move melody pointer
		{
			sound_duration--;							// -1 millisecond
			// sound melody pointer movement when current tone expired while alarming?
			if( !sound_duration && bv(RTC_Alarming,BAPP1) )
			{
				if(sound_melody_idx >= MELODY_LENGTH)
				{
					sound_melody_idx=0; 				// from the top
				}

				sound_freq = (uint8_t)pgm_read_word(melody + sound_melody_idx);
				sound_melody_idx++;
				sound_duration = pgm_read_word(melody + sound_melody_idx);
				sound_melody_idx++;
			}
			// check for sound end?
			if(!sound_duration)							// check for end, to shut down or to move melody pointer
			{
				SOUND_PORT &= ~_BV(SOUND_PIN);			// turn it off
				sound_melody_idx = 0;					// set for next alarm
			}
		}
	}

	// has to buzz?
	if(sound_duration)
	{
		// need to sound or make a delay (>0)
		if(sound_freq)
		{
			if(++sound_presc >= sound_freq)					// next, &check for overflow
			{
				sound_presc = 0;
				SOUND_PORT ^= _BV(SOUND_PIN);				// toggling this produces the sound on speaker
			}
		}
		// this is delay between sounding, turn the buzzer off while delaying
		else
		{
			SOUND_PORT &= ~_BV(SOUND_PIN);					// turn (keep) it off during tone-pause
		}
	}
}


//##########################
// Interrupt: PCINT1_vect //
//##########################
// called on PCINT1_vect - pin change interrupt
ISR(PCINT1_vect, ISR_NOBLOCK) // this interrupt routine can be interrupted by any other, and MUST BE!
{
	//sei(); // ! this must BE, (or with ISR_NOBLOCK) because we are relying on delay_ms_() which is actually an interrupt driven delay function

	// Info: since this is a non-blocking interrupt, it might be possible that we re-enter this ISR while button debounces! that might cause problems, but... (see PCINT0 vector)
	PCICR &= ~_BV(PCIE1); 								// ...but we will do it anyway.. and this is how

	// check if PWR-SENS is connected to GND (absence of VCC)
	if( !(PWR_SENS_PINREG & _BV(PWR_SENS_PIN)) )
	{
		// if it was not GND before, and now is
		if ( !(prev_btn & 0b10000000) )
		{
			delay_ms_(15); 								// debounce to see if it was just a power surge
			if( PWR_SENS_PINREG & _BV(PWR_SENS_PIN) )	// if it is still not GND after 50ms, this was an error
			{
				PCICR |= _BV(PCIE1); 					// ..re-enable interrupts for the entire section
				return;
			}
			prev_btn |= 0b10000000; 			// it really is GND now, mark it so
			sleep_go(); 								// goto sleep
		}
	}
	// the power is here!
	else
	{
		// if it was GND before, and isn't anymore
		if ( prev_btn & 0b10000000 )
		{
			delay_ms_(15); 								// debounce to see if it was just a power flicker
			if( !(PWR_SENS_PINREG & _BV(PWR_SENS_PIN)) )// if it is still not VCC after 50ms, this was an error
			{
				PCICR |= _BV(PCIE1); 					// ..re-enable interrupts for the entire section
				return;
			}
			prev_btn &= ~0b10000000; 					// say that it is not GND anymore
			sleep_end(); 								// end sleep, goto online
		}
	}

	// check to see if button A is pressed
	if( !(BTNA_PINREG & _BV(BTNA_PIN)) )
	{
		// see if it wasn't pressed at all
		if ( !(prev_btn & 0b00000001) )
		{
			delay_ms_(25); // debounce
			if( BTNA_PINREG & _BV(BTNA_PIN) )			// if it is still not GND after 25ms, this was an error
			{
				PCICR |= _BV(PCIE1); 					// ..re-enable interrupts for the entire section
				return;
			}
			prev_btn |= 0b00000001; 					// remember it was pressed

			// if we are in auto-off then we will go online for 30sec, do the job, and go back to auto-off mode
			if( bv(AutoOffed,BSYS) )
			{
				auto_off_auto_executer=30; 				// enable auto-off auto-executer
				bc(AutoOffed,BSYS);
				boost_continue(); 						// continue the booster
			}

			btn_hold_timer = BTN_HOLD_TMR; 				// load for hold detection
			while(btn_hold_timer > 0)
			{
				// button released in the meantime
				if( BTNA_PINREG & _BV(BTNA_PIN) )
				{
					btn_press |= 0b00000001; 			// this was just a press!

					// snooze override
					if( bv(RTC_Alarming,BAPP1) )
					{
						btn_hold &= ~0b00000001; 		// clear the key event
						btn_press &= ~0b00000001; 	// clear the key event
						alarm_snooze();
					}
					PCICR |= _BV(PCIE1); 				// ..re-enable interrupts for the entire section
					return;
				}
			}
			// if we get to this line, the button is still pressed, so "it is held"
			btn_hold |= 0b00000001;
		}
	}
	// button not pressed at this ISR event
	else
	{
		prev_btn &= ~0b00000001;
	}

	// check to see if button B is pressed
	if( !(BTNB_PINREG & _BV(BTNB_PIN)) )
	{
		// see if it wasn't pressed at all
		if ( !(prev_btn & 0b00000010) )
		{
			delay_ms_(25); // debounce
			if( BTNB_PINREG & _BV(BTNB_PIN) )			// if it is still not GND after 25ms, this was an error
			{
				PCICR |= _BV(PCIE1); 					// ..re-enable interrupts for the entire section
				return;
			}
			prev_btn |= 0b00000010; 					// remember it was pressed

			// if we are in auto-off then we will go online for 30sec, do the job, and go back to auto-off mode
			if( bv(AutoOffed,BSYS) )
			{
				auto_off_auto_executer=30; 				// enable auto-off auto-executer
				bc(AutoOffed,BSYS);
				boost_continue(); // continue the booster
			}

			btn_hold_timer = BTN_HOLD_TMR; 				// load for hold detection
			while(btn_hold_timer > 0)
			{
				// button released in the meantime
				if( BTNB_PINREG & _BV(BTNB_PIN) )
				{
					btn_press |= 0b00000010; 			// this was just a press!

					// snooze override
					if( bv(RTC_Alarming,BAPP1) )
					{
						btn_hold &= ~0b00000010; 		// clear the key event
						btn_press &= ~0b00000010; 	// clear the key event
						alarm_snooze();
					}
					PCICR |= _BV(PCIE1); 				// ..re-enable interrupts for the entire section
					return;
				}
			}
			// if we get to this line, the button is still pressed, so "it is held"
			btn_hold |= 0b00000010;
		}
	}
	// button not pressed at this ISR event
	else
	{
		prev_btn &= ~0b00000010;
	}

	PCICR |= _BV(PCIE1); 								// ..re-enable interrupts for the entire section

	return;
}


//##########################
// Interrupt: PCINT0_vect //
//##########################
// called on PCINT0_vect - pin change interrupt
ISR(PCINT0_vect, ISR_NOBLOCK) // this interrupt routine can be interrupted by any other, and MUST BE!
{
	//sei(); // ! this must BE, (or with ISR_NOBLOCK) because we are relying on delay_ms_() which is actually an interrupt driven delay function

	//SWALARM_PCMSKREG &= ~_BV(SWALARM_PCINTBIT);				// alarm switch interrupt disabled - because this is a non-blocking interrupt and it will go in here while switch debounces
	PCICR &= ~_BV(PCIE0); 										// ..this is how we will do it

	// dummy do..while, just for "break"-ing and re-enabling alarm switch interrupts :)
	do
	{
		// handle the alarm-switch, alarm on
		if( !(SWALARM_PINREG & _BV(SWALARM_PIN)) )
		{
			// if it was not GND before, and now is
			if ( !(prev_btn & 0b00000100) )
			{
				delay_ms_(50); 									// debounce
				if( SWALARM_PINREG & _BV(SWALARM_PIN) ) break;	// if it is not still GND after 50ms, this was an error
				prev_btn |= 0b00000100; 						// it really is GND now, mark it so

				alarm_off();									// ALARM IS NOW OFF!
			}
		}
		// alarm off
		else
		{
			// if it was GND before, and isn't anymore
			if ( prev_btn & 0b00000100 )
			{
				delay_ms_(50); 									// debounce
				if( !(SWALARM_PINREG & _BV(SWALARM_PIN)) ) break;// if it is not still VCC after 50ms, this was an error
				prev_btn &= ~0b00000100; 						// say that it is not GND anymore

				alarm_on();										// ALARM IS NOW ON!
			}
		}
	} while(0); // will not iterate

	//SWALARM_PCMSKREG |= _BV(SWALARM_PCINTBIT);				// alarm switch re-enabled
	PCICR |= _BV(PCIE0); 										// ..this is how we will do it

	return;
}


//###################
// Misc. functions //
//###################
// fills-in display_digits[]s with HH:MM
void display_time()
{
	bs(ToggleColons,BAPP1); 								// toggle colons because we are displaying time

	// time not set? .. will be toggled every RTC[TIME_S]%2
	if( bv(BlankTime,BAPP1) )
	{
		display_digits[0]=display_digits[1]=display_digits[2]=display_digits[3]=nixie_blank;
		return;
	}

	uint8_t display_digits_temp[DISPLAY_SIZE];
	display_digits_temp[3] = (RTC[TIME_M] % 10); 			// HH  :M[M]
	display_digits_temp[2] = (RTC[TIME_M] / 10); 			// HH  :[M]M
	display_digits_temp[1] = ((RTC[TIME_H] % 24) % 10); 	// H[H]:MM
	display_digits_temp[0] = ((RTC[TIME_H] % 24) / 10); 	// [H]H:MM

	// anti-poison thing - mostly because it looks great
	if( bv(AntiPoison,BAPP1) )
	{
		bc(CrossFaderEn,BAPP2);								// and this
		uint8_t i;
		uint8_t j;

		for(j=0; j<DISPLAY_SIZE; j++)
		{
			for(i=0; i<10; i++)
			{
				display_digits[j] = i;
				delay_ms_(20);
			}
			display_digits[j] = display_digits_temp[j];
		}

		bc(AntiPoison,BAPP1); 								// clear the anti-poison bit
		if( bv(CrossFading,BUSR1) ) bs(CrossFaderEn,BAPP2);	// restore this
	}

	display_digits[3] = display_digits_temp[3];
	display_digits[2] = display_digits_temp[2];
	display_digits[1] = display_digits_temp[1];
	display_digits[0] = display_digits_temp[0];

	/*
	// all code above can be reduced to this below, if anti-poison is not used
	display_digits[3] = (RTC[TIME_M] % 10); 				// HH  :M[M]
	display_digits[2] = (RTC[TIME_M] / 10); 				// HH  :[M]M
	display_digits[1] = ((RTC[TIME_H] % 24) % 10); 			// H[H]:MM
	display_digits[0] = ((RTC[TIME_H] % 24) / 10); 			// [H]H:MM
	*/

	return;
}
// reads DS18S20's temperature and fills-in display_digits[]s with DD'D and holds there for duration parameter
void display_temp(uint16_t duration)
{
	int8_t temp = 0;

	temp = therm_read_temperature(); 					// get the temperature
	if(temp != THERM_INVALID) 							// refresh the display if received temperature is valid
	{
		bc(ToggleColons,BAPP1); 						// we must clear this baby
		bc(ToggleLEDs,BAPP1); 							// and this, since we use the fourth LED to mark negative temperatures
		bc(CrossFaderEn,BAPP2);							// and this

		set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN); 		// lower dot off
		set1(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN); 		// upper dot on (degrees mark)

		// do a little conversion now...
		if(temp & _BV(0)) 								// if the temperature has .5 degree
			display_digits[2] = 5;
		else
			display_digits[2] = 0; 						// or "nixie_blank" if you prefer

		temp /= 2; 										// convert from 0.5 steps to 1 step

		// Lets get RGB LED values for current temperature, if the use wants
		if( bv(TEMPblender,BUSR1) )
		{
			uint8_t Rval = rgb_blender(Rmax,Rmin,temp);
			uint8_t Gval = rgb_blender(Gmax,Gmin,temp);
			uint8_t Bval = rgb_blender(Bmax,Bmin,temp);
			rgb_led(0,Rval,Gval,Bval);
			rgb_led(1,Rval,Gval,Bval);
			rgb_led(2,Rval,Gval,Bval);
		}
		else
		{
			rgb_led(0,0,0,0);
			rgb_led(1,0,0,0);
			rgb_led(2,0,0,0);
		}

		if(temp < 0)
		{
			rgb_led(3,0,0,255); 						// we don't have the minus sign, so we will color fourht nixie with blue LED for negative temperatures
		}
		else
		{
			rgb_led(3,0,0,0);
		}

		display_digits[0] = (temp / 10);
		display_digits[1] = (temp % 10);

		display_digits[3] = nixie_blank; 				// this one is always off

		delay_ms_(duration);							// stay in here for wanted period of time

		// activate user preferences to application
		//restore_setup(); // moved this at the location after calling display_temp() function!
	}

	return;
}
// displays full date on 4 digits
void display_date()
{
	bc(ToggleColons,BAPP1); 							// don't need this
	bc(ToggleLEDs,BAPP1); 								// this is bad also
	bc(CrossFaderEn,BAPP2);								// and this

	// bottom dot only
	set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
	set1(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);

	// dd/mm
	display_digits[0] = (RTC[DATE_D] / 10);				// [D]D/MM
	display_digits[1] = (RTC[DATE_D] % 10);				// D[D]/MM
	display_digits[2] = (RTC[DATE_M] / 10);				// DD  /[M]M
	display_digits[3] = (RTC[DATE_M] % 10);				// DD  /M[M]

	// Dim the MM part if dimming enabled
	if( bv(DateFading,BUSR1) )
	{
		brightness[0] = brightness[1] = max_brightness; 	// back to highest
		brightness[2] = brightness[3] = date_fading_brightness; // dimmed
	}
	// use leds?
	if( bv(LEDs4date,BUSR1) )
	{
		rgb_led(0,rgb_date_color[0],rgb_date_color[1],rgb_date_color[2]);
		rgb_led(1,rgb_date_color[0],rgb_date_color[1],rgb_date_color[2]);
		rgb_led(2,0,0,0);
		rgb_led(3,0,0,0);
	}

	delay_ms_(SHOW_DATE_DDMM_DLY/2);

	// Just dim the DD part now if dimming enabled
	if( bv(DateFading,BUSR1) )
	{
		brightness[0] = brightness[1] = date_fading_brightness; // dimmed
		brightness[2] = brightness[3] = max_brightness; 	// back to highest
	}
	// use leds?
	if( bv(LEDs4date,BUSR1) )
	{
		rgb_led(0,0,0,0);
		rgb_led(1,0,0,0);
		rgb_led(2,rgb_date_color[0],rgb_date_color[1],rgb_date_color[2]);
		rgb_led(3,rgb_date_color[0],rgb_date_color[1],rgb_date_color[2]);
	}

	delay_ms_(SHOW_DATE_DDMM_DLY/2);

	// both dots off
	set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
	set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);

	// yyyy
	display_digits[0] = 2;								// [Y]YYY
	display_digits[1] = (RTC[DATE_Y] / 100);			// Y[Y]YY
	display_digits[2] = ((RTC[DATE_Y] % 100) / 10);		// YY[Y]Y
	display_digits[3] = ((RTC[DATE_Y] % 100) % 10);		// YYY[Y]

	// Dim all four if dimming enabled
	if( bv(DateFading,BUSR1) )
	{
		brightness[0] = brightness[1] = brightness[2] = brightness[3] = date_fading_brightness; // all dimmed
	}
	// use leds?
	if( bv(LEDs4date,BUSR1) )
	{
		rgb_led(0,rgb_date_color[0],rgb_date_color[1],rgb_date_color[2]);
		rgb_led(1,rgb_date_color[0],rgb_date_color[1],rgb_date_color[2]);
		//rgb_led(2,rgb_date_color[0],rgb_date_color[1],rgb_date_color[2]);	// optimisation
		//rgb_led(3,rgb_date_color[0],rgb_date_color[1],rgb_date_color[2]);	// optimisation
	}

	delay_ms_(SHOW_DATE_YYYY_DLY);

	// restore brightness to maximum if dimming was used
	if( bv(DateFading,BUSR1) )
	{
		set_brightness_of(DISPLAY_SIZE,max_brightness);		// set the brightness of all tubes back to max
	}

	// use leds?
	if( bv(LEDs4date,BUSR1) )
	{
		rgb_led(0,0,0,0);
		rgb_led(1,0,0,0);
		rgb_led(2,0,0,0);
		rgb_led(3,0,0,0);
	}

	// activate user preferences to application
	restore_setup();

	return;
}
// display menu for setup
void display_menu()
{
	/*
		0 1 x x = set time&date
		0 2 x x = set alarm time
		0 3 x x = set snooze interval in minutes
		0 4 x x = set mode times
		0 5 x x = set brightness
		0 6 x x = set time led preferences (1/2Hz toggling and color)
		0 7 x x = set temperature blender on/off
		0 8 x x = set date led preferences (on/off and color)
		0 9 x x = set date fading (on/off)
		1 0 x x = set auto-off times and on/off
		1 1 x x = set cross-fader speed and on/off

		n n x x = exit menu (the last option +1)
	*/
	uint8_t option_id = 1;

	bc(ToggleColons,BAPP1); 							// don't need this
	bc(ToggleLEDs,BAPP1); 								// this is bad also
	bc(CrossFaderEn,BAPP2);							// and this

	btn_inactivity_timer = BTN_INACTIVITY_TMR;			// if the user forgets to click around

	while(1)
	{
		if(btn_inactivity_timer==0) break;				// ending the "menu"

		// both dots off
		set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
		set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);
		rgb_led(0,0,0,255);								// color these blue
		rgb_led(1,0,0,255);
		rgb_led(2,0,0,0);								// off us
		rgb_led(3,0,0,0);

		display_digits[0] = option_id / 10;				// show currently selected value
		display_digits[1] = option_id % 10;				// in XX format... 01-99 options :)
		display_digits[2] = display_digits[3] = nixie_blank; // these are blank

		// BTN-A pressed or held?
		if((btn_press|btn_hold) & 0b00000001)
		{
			btn_press &= ~0b00000001; 				// button handled
			btn_hold &= ~0b00000001;					// also handle the hold
			if(++option_id > 11+1) option_id=1;			// increment option, check for end, allow +1, that will be the "exit menu"
			sound_beep(100,5);							// tick
			while( !(BTNA_PINREG & _BV(BTNA_PIN)) );	// stay here while it is pressed
			delay_ms_(30);
			btn_inactivity_timer = BTN_INACTIVITY_TMR;	// reload
		}
		// BTN-B pressed or held?
		if((btn_press|btn_hold) & 0b00000010)
		{
			btn_press &= ~0b00000010; 				// button handled
			btn_hold &= ~0b00000010;					// also handle the hold
			sound_beep(100,5);							// tick
			sound_beep(20,15);							// tick
			while( !(BTNB_PINREG & _BV(BTNB_PIN)) );	// stay here while it is pressed
			delay_ms_(30);
			btn_inactivity_timer = BTN_INACTIVITY_TMR;	// reload

			switch(option_id)							// what was the user's choice?
			{
				// set time&date
				case 1:
					set_rtc();
				break;

				// set alarm time
				case 2:
					set_alarm_time();
				break;

				// set snooze interval
				case 3:
					set_snooze_interval();
				break;

				// set mode times
				case 4:
					set_mode_times();
				break;

				// set brightness
				case 5:
					set_brightness();
				break;

				// set bottom led preferences for clock
				case 6:
					set_time_led_prefs();
				break;

				// set temperature blender on/off
				case 7:
					set_temp_blender();
				break;

				// set bottom led preferences for date
				case 8:
					set_date_led_prefs();
				break;

				// set fading value and on/off for nixie tubes in date display?
				case 9:
					set_date_fading();
				break;

				// set auto-off times and on/off
				case 10:
					set_auto_off_prefs();
				break;

				// set cross-fading speed and on/off
				case 11:
					set_cross_fader_prefs();
				break;

				default:
					btn_inactivity_timer=0; // this is how we exit the menu (or we wait for the timeout)
			}
		}
	}

	// save settings to eeprom
	uint8_t *pEE = (uint8_t *)EEPROM_START;
	eeprom_write_byte(pEE++,0xAA); // say that the EEPROM is valid
	eeprom_write_block((const void *)&RTC,pEE,6);
	pEE+=6;
	eeprom_write_block((const void *)&RTC_ALARM,pEE,2);
	pEE+=2;
	eeprom_write_block((const void *)&RTC_AOFF,pEE,2);
	pEE+=2;
	eeprom_write_block((const void *)&RTC_AON,pEE,2);
	pEE+=2;
	eeprom_write_block((const void *)&rgb_time_color,pEE,3);
	pEE+=3;
	eeprom_write_block((const void *)&rgb_date_color,pEE,3);
	pEE+=3;
	eeprom_write_word((uint16_t *)pEE,max_brightness);
	pEE+=2;
	eeprom_write_word((uint16_t *)pEE,date_fading_brightness);
	pEE+=2;
	eeprom_write_byte(pEE++,snooze_interval);
	eeprom_write_block((const void *)&disp_duration,pEE,DISP_THINGS);
	pEE+=DISP_THINGS;
	/*
	eeprom_write_block((const void *)&bools,pEE,BOOL_BANK_SIZE);
	*/
	eeprom_write_byte(pEE,bools[BUSR1]);

	// end-of-programming info:
	sound_beep(100,5);								// tick
	sound_beep(20,15);								// tick
	sound_beep(0,10);								// wait...
	sound_beep(100,5);								// tick
	sound_beep(20,15);								// tick

	faded_chase_leds(0,0,0, 20,250,0, 10, 1000, 1);	// lets try faded chasing GREEN for fun
	faded_chase_leds(0,0,0, 20,0,250, 10, 1000, 2);	// lets try faded chasing GREEN for fun, backwards

	disp_duration_timer = 0;						// start over after exiting the programming mode
	disp_what = 0;									// ?

	// activate user preferences to application
	restore_setup();

	return;												// return from menu and go back to "idle" mode in the main()
}
// setting the time and date
void set_rtc()
{
	uint8_t setup_part = 0; 							// first we are going to setup the HH

	btn_inactivity_timer = BTN_INACTIVITY_TMR;			// if the user ignores the clock

	while(setup_part < 5)								// we are setting 5 parts in total, from 0..4 (HH, MM, dd, mm, yyyy)
	{
		RTC[TIME_S] = 0;								// so that RTC doesn't screw with us while we set the time, and to start from 0 on exit

		if(btn_inactivity_timer==0) break;				// get out at the exit point of this function

		// time parts: 0, 1
		if(setup_part <= 1)
		{
			// both dots on
			set1(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
			set1(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);
			if(setup_part == 0)
			{
				rgb_led(0,0,0,255);						// blue us up
				rgb_led(1,0,0,255);
				rgb_led(2,0,0,0);						// off us
				rgb_led(3,0,0,0);
			}
			else
			{
				rgb_led(0,0,0,0);						// off us
				rgb_led(1,0,0,0);
				rgb_led(2,0,0,255);						// blue us up
				rgb_led(3,0,0,255);
			}
			// show the value: hh:mm
			display_digits[0]=((RTC[TIME_H] % 24) / 10);// [H]H:MM
			display_digits[1]=((RTC[TIME_H] % 24) % 10);// H[H]:MM
			display_digits[2]=(RTC[TIME_M] / 10);		// HH  :[M]M
			display_digits[3]=(RTC[TIME_M] % 10);		// HH  :M[M]
		}
		// date parts: 2, 3, 4
		else if( setup_part <= 4)
		{
			// if we are setting the year part
			if(setup_part == 4)
			{
				// both dots off
				set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
				set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);

				// show the value: yyyy
				display_digits[0] = 2;					// [Y]YYY
				display_digits[1] = (RTC[DATE_Y] / 100);// Y[Y]YY
				display_digits[2] = ((RTC[DATE_Y] % 100) / 10);// YY[Y]Y
				display_digits[3] = ((RTC[DATE_Y] % 100) % 10);// YYY[Y]

				rgb_led(0,0,0,255);							// also blue up the first two, last two are already blue
				rgb_led(1,0,0,255);
			}
			// nope, it is still the DD. or .MM part
			else // 2 or 3
			{
				// bottom dot on
				set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
				set1(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);
				if(setup_part == 2)
				{
					rgb_led(0,0,0,255);						// blue us up
					rgb_led(1,0,0,255);
					rgb_led(2,0,0,0);						// off us
					rgb_led(3,0,0,0);
				}
				else //3
				{
					rgb_led(0,0,0,0);						// off us
					rgb_led(1,0,0,0);
					rgb_led(2,0,0,255);						// blue us up
					rgb_led(3,0,0,255);
				}

				// show the value: dd.mm
				display_digits[0] = (RTC[DATE_D] / 10);		// [D]D/MM
				display_digits[1] = (RTC[DATE_D] % 10);		// D[D]/MM
				display_digits[2] = (RTC[DATE_M] / 10);		// DD  /[M]M
				display_digits[3] = (RTC[DATE_M] % 10);		// DD  /M[M]
			}
		}

		// BTN-A? held or pressed
		if( (btn_press|btn_hold) & 0b00000001)
		{
			uint16_t delay_value = BTN_REPEAT_RATE;			// speed of auto-increment when holding a button
			switch(setup_part)
			{
				// the HH:mm dd.mm.yyyy
				case 0:
					RTC[TIME_H] = (RTC[TIME_H]+1) % 24;		// +1
				break;
				// the hh:MM dd.mm.yyyy
				case 1:
					RTC[TIME_M] = (RTC[TIME_M]+1) % 60;		// +1
				break;
				// the hh:mm DD.mm.yyyy
				case 2:
					if(++RTC[DATE_D] > 31) RTC[DATE_D] = 1;	// +1 with check
				break;
				// the hh:mm dd.MM.yyyy
				case 3:
					if(++RTC[DATE_M] > 12) RTC[DATE_M] = 1;	// +1 with check
				break;
				// the hh:mm dd.mm.yYYY
				case 4:
					RTC[DATE_Y] = RTC[DATE_Y]+1;			// +1 .. might be boring to wait for overflow at year 2256 when setting :)
					delay_value = BTN_REPEAT_RATE/2;		// so this will go twice as fast!
				break;
			}

			if(btn_hold & 0b00000001)						// if the user is holding it, increment hour by 1s
			{
				delay_ms_(delay_value);						// slow down but not too much
			}

			btn_press &= ~0b00000001; 					// button handled
			if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
			sound_beep(100,5);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload
		}

		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 					// button handled
			btn_hold &= ~0b00000010;						// also handle the hold
			setup_part++;
			sound_beep(100,5);								// tick
			sound_beep(20,15);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload to stay in programming mode
		}

	} // while(setup_part < 5)

	// we have just done it
	bs(RTC_IsSet,BAPP1);
	bc(BlankTime,BAPP1);									// in case we entered programming while the display was blanked

	return;
}
// setting the time of alarm
void set_alarm_time()
{
	uint8_t setup_part = 0; 								// first we are going to setup the HH

	btn_inactivity_timer = BTN_INACTIVITY_TMR;				// if the user forgets to click around

	// both dots on
	set1(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
	set1(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);

	while(setup_part < 2)									// setting only HH=0 and MM=1 parts
	{
		if(btn_inactivity_timer==0) break;					// get out at the exit point of this function

		if(setup_part == 0)
		{
			rgb_led(0,0,0,255);								// blue us up
			rgb_led(1,0,0,255);
			rgb_led(2,0,0,0);								// off us
			rgb_led(3,0,0,0);
		}
		else
		{
			rgb_led(0,0,0,0);								// off us
			rgb_led(1,0,0,0);
			rgb_led(2,0,0,255);								// blue us up
			rgb_led(3,0,0,255);
		}
		// show the value: hh:mm
		display_digits[0]=((RTC_ALARM[TIME_H] % 24) / 10);	// [H]H:MM
		display_digits[1]=((RTC_ALARM[TIME_H] % 24) % 10);	// H[H]:MM
		display_digits[2]=(RTC_ALARM[TIME_M] / 10);			// HH  :[M]M
		display_digits[3]=(RTC_ALARM[TIME_M] % 10);			// HH  :M[M]

		// BTN-A? held or pressed
		if( (btn_press|btn_hold) & 0b00000001)
		{
			switch(setup_part)
			{
				// the HH:mm
				case 0:
					RTC_ALARM[TIME_H] = (RTC_ALARM[TIME_H]+1) % 24; // +1
				break;
				// the hh:MM
				case 1:
					RTC_ALARM[TIME_M] = (RTC_ALARM[TIME_M]+1) % 60; // +1
				break;
			}

			if(btn_hold & 0b00000001)							// if the user is holding it, increment hour by 1s
			{
				delay_ms_(BTN_REPEAT_RATE);						// slow down but not too much
			}

			btn_press &= ~0b00000001; 						// button handled
			if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
			sound_beep(100,5);									// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 			// reload
		}

		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 						// button handled
			btn_hold &= ~0b00000010;							// also handle the hold
			setup_part++;
			sound_beep(100,5);									// tick
			sound_beep(20,15);									// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 			// reload to stay in programming mode
		}

	}

	return;
}
// set snooze interval in minutes
void set_snooze_interval()
{
	rgb_led(0,0,0,0);										// off with the first two
	rgb_led(1,0,0,0);
	rgb_led(2,0,0,255);										// blue these - the value we choose
	rgb_led(3,0,0,255);

	btn_inactivity_timer = BTN_INACTIVITY_TMR;				// if the user forgets to click around

	while(1)
	{
		if(btn_inactivity_timer==0) break;					// get out at the exit point of this function

		// current brightness value: x x Y Y
		display_digits[0] = display_digits[1] = nixie_blank;// blank first two digits
		display_digits[2] = (snooze_interval / 10); 		// tens
		display_digits[3] = (snooze_interval % 10); 		// ones

		// BTN-A? held or pressed
		if( (btn_press|btn_hold) & 0b00000001)
		{
			snooze_interval++;								// increment it!
			if(snooze_interval>99) snooze_interval=(RTC_AUTO_SNOOZE*2)/60;// rollover to min allowed and that is twice the alarming period but in minutes

			if(btn_hold & 0b00000001)						// if the user is holding it, increment duration by 1s
			{
				delay_ms_(BTN_REPEAT_RATE);					// slow down
			}

			btn_press &= ~0b00000001; 					// button handled
			if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
			sound_beep(100,5);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload
		}

		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 					// button handled
			btn_hold &= ~0b00000010;						// also handle the hold
			sound_beep(100,5);								// tick
			sound_beep(20,15);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload to stay in programming mode
			break;											// exit at the exit point of this function
		}
	}

	return;
}
// setting mode durations
void set_mode_times()
{
	rgb_led(0,0,0,255);										// blue the option parts
	rgb_led(1,0,0,0);
	rgb_led(2,0,255,0);										// green the values
	rgb_led(3,0,255,0);

	btn_inactivity_timer = BTN_INACTIVITY_TMR;				// if the user forgets to click around

	uint8_t setup_part = 0;
	while(setup_part < DISP_THINGS)
	{
		if(btn_inactivity_timer==0) break;					// get out at the exit point of this function

		// current option in format X_:YY
		display_digits[0] = (setup_part+1); 				// 0 will be 1, and so on...
		display_digits[1] = nixie_blank;
		// current value in YY format
		display_digits[2] = disp_duration[setup_part] / 10;
		display_digits[3] = disp_duration[setup_part] % 10;

		// BTN-A? held or pressed
		if( (btn_press|btn_hold) & 0b00000001)
		{
			disp_duration[setup_part]++;					// increment it!
			if(disp_duration[setup_part]>99) disp_duration[setup_part]=0; // limit to 99
			if(btn_hold & 0b00000001)						// if the user is holding it, increment duration by 1s
			{
				delay_ms_(BTN_REPEAT_RATE);					// slow down
			}

			btn_press &= ~0b00000001; 					// button handled
			if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
			sound_beep(100,5);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload
		}

		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 					// button handled
			btn_hold &= ~0b00000010;						// also handle the hold
			setup_part++;
			sound_beep(100,5);								// tick
			sound_beep(20,15);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload to stay in programming mode
		}
	}

	return;
}
// set display brightness
void set_brightness()
{
	rgb_led(0,0,0,255);										// blue all
	rgb_led(1,0,0,255);
	rgb_led(2,0,0,255);
	rgb_led(3,0,0,255);

	btn_inactivity_timer = BTN_INACTIVITY_TMR;				// if the user forgets to click around

	while(1)
	{
		if(btn_inactivity_timer==0) break;					// get out at the exit point of this function

		// current brightness value: X X X X
		display_digits[0] = (max_brightness / 1000);				// thousands
		display_digits[1] = ((max_brightness % 1000) / 100); 		// hundreds
		display_digits[2] = (((max_brightness % 1000) % 100) / 10); // tens
		display_digits[3] = (((max_brightness % 1000) % 100) % 10); // ones

		// BTN-A? held or pressed
		if( (btn_press|btn_hold) & 0b00000001)
		{
			if(btn_hold & 0b00000001)						// if the user is holding it, increment duration by 1s
			{
				// go slower after 4000 :)
				if(max_brightness>=4000)
				{
					delay_ms_(BTN_REPEAT_RATE);				// slow down
					max_brightness+=5;						// increment it!
				}
				// else go very fast
				else
				{
					delay_ms_(BTN_REPEAT_RATE/2);			// slow down, but go faster than usual
					max_brightness+=50;						// increment it!
				}
			}
			else
			{
				max_brightness++;							// increment it!
			}
			set_brightness_of(DISPLAY_SIZE,max_brightness);	// set all to this

			// check the maximum allowed value
			if(max_brightness>DISP_MAX_BRIGHTNESS) max_brightness=DISP_MIN_BRIGHTNESS; // rollover to min allowed

			// check the minimum allowed value: because cross-fading doesn't work when user's set brightness is less than "DISP_LAST_OFF_BRTNS" constant don't let it be
			if(max_brightness<DISP_MIN_BRIGHTNESS) max_brightness=DISP_MIN_BRIGHTNESS; // again, rollover to min allowed -- was <DISP_LAST_OFF_BRTNS; 02-11-2010

			btn_press &= ~0b00000001; 					// button handled
			if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
			sound_beep(100,5);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload
		}

		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 					// button handled
			btn_hold &= ~0b00000010;						// also handle the hold
			sound_beep(100,5);								// tick
			sound_beep(20,15);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload to stay in programming mode
			break;											// exit brightness programming at the exit point of this function
		}
	}

	return;
}
// set bottom leds toggling and their color for TIME
void set_time_led_prefs(void)
{
	uint8_t setup_part = 0; 								// first we are going to setup the HH

	btn_inactivity_timer = BTN_INACTIVITY_TMR;				// if the user forgets to click around

	while(setup_part < 4)									// 3 colors, and on/off = 4 things to set in total
	{
		if(btn_inactivity_timer==0) break;					// get out at the exit point of this function

		// show somehow which color is being set now (R,G,B)
		display_digits[0] = ( setup_part==0 ? 1 : nixie_blank );
		display_digits[1] = ( setup_part==1 ? 1 : nixie_blank );
		display_digits[2] = ( setup_part==2 ? 1 : nixie_blank );
		// put current on/off setting on last nixie (1/0)
		if(setup_part==3)
		{
			display_digits[3] = ( bv(LEDtoggling,BUSR1) ? 1:0 );
		}
		else
		{
			display_digits[3] = nixie_blank;
		}

		rgb_led(0,rgb_time_color[0],0,0);					// show red here
		rgb_led(1,0,rgb_time_color[1],0);					// show green here
		rgb_led(2,0,0,rgb_time_color[2]);					// show blue here
		rgb_led(3,rgb_time_color[0],rgb_time_color[1],rgb_time_color[2]);	// show all colors mixed here

		// BTN-A? held or pressed
		if( (btn_press|btn_hold) & 0b00000001)
		{
			// led colors are being set
			if(setup_part < 3)
			{
				rgb_time_color[setup_part]++;				// increment current color value! it will overflow after 255 to 0
			}
			// on/off is being set now
			else
			{
				bt(LEDtoggling,BUSR1);						// toggle between 1 and 0
			}

			if(btn_hold & 0b00000001)						// if the user is holding it, increment duration by 1s
			{
				delay_ms_(BTN_REPEAT_RATE/2);				// slow down, but repeat faster than usual
			}

			btn_press &= ~0b00000001; 					// button handled
			if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
			sound_beep(100,5);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload
		}

		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 					// button handled
			btn_hold &= ~0b00000010;						// also handle the hold
			setup_part++;									// next part
			sound_beep(100,5);								// tick
			sound_beep(20,15);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload to stay in programming mode
		}
	}

	return;
}
// set temperature color blender on/off
void set_temp_blender(void)
{
	rgb_led(0,0,0,0);										// off with the first two
	rgb_led(1,0,0,0);
	rgb_led(2,0,0,0);
	rgb_led(3,0,0,255);										// blue this one

	btn_inactivity_timer = BTN_INACTIVITY_TMR;				// if the user forgets to click around

	while(1)
	{
		if(btn_inactivity_timer==0) break;					// get out at the exit point of this function

		// current brightness value: x x Y Y
		display_digits[0] = display_digits[1] = display_digits[2] = nixie_blank;// blank first 3 digits
		// put current on/off setting on last nixie (1/0)
		display_digits[3] = ( bv(TEMPblender,BUSR1) ? 1:0 );

		// BTN-A? held or pressed
		if( (btn_press|btn_hold) & 0b00000001)
		{
			bt(TEMPblender,BUSR1);							// toggle between 1 and 0

			if(btn_hold & 0b00000001)						// if the user is holding it, increment duration by 1s
			{
				delay_ms_(BTN_REPEAT_RATE);					// slow down
			}

			btn_press &= ~0b00000001; 					// button handled
			if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
			sound_beep(100,5);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload
		}

		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 					// button handled
			btn_hold &= ~0b00000010;						// also handle the hold
			sound_beep(100,5);								// tick
			sound_beep(20,15);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload to stay in programming mode
			break;											// exit at the exit point of this function
		}
	}

	return;
}
// set bottom leds on/off and their color for DATE
void set_date_led_prefs(void)
{
	uint8_t setup_part = 0; 								// first we are going to setup the HH

	btn_inactivity_timer = BTN_INACTIVITY_TMR;				// if the user forgets to click around

	while(setup_part < 4)									// 3 colors, and on/off = 4 things to set in total
	{
		if(btn_inactivity_timer==0) break;					// get out at the exit point of this function

		// show somehow which color is being set now (R,G,B)
		display_digits[0] = ( setup_part==0 ? 1 : nixie_blank );
		display_digits[1] = ( setup_part==1 ? 1 : nixie_blank );
		display_digits[2] = ( setup_part==2 ? 1 : nixie_blank );
		// put current on/off setting on last nixie (1/0)
		if(setup_part==3)
		{
			display_digits[3] = ( bv(LEDs4date,BUSR1) ? 1:0 );
		}
		else
		{
			display_digits[3] = nixie_blank;
		}

		rgb_led(0,rgb_date_color[0],0,0);					// show red here
		rgb_led(1,0,rgb_date_color[1],0);					// show green here
		rgb_led(2,0,0,rgb_date_color[2]);					// show blue here
		rgb_led(3,rgb_date_color[0],rgb_date_color[1],rgb_date_color[2]);	// show all colors mixed here

		// BTN-A? held or pressed
		if( (btn_press|btn_hold) & 0b00000001)
		{
			// led colors are being set
			if(setup_part < 3)
			{
				rgb_date_color[setup_part]++;				// increment current color value! it will overflow after 255 to 0
			}
			// on/off is being set now
			else
			{
				bt(LEDs4date,BUSR1);						// toggle between 1 and 0
			}

			if(btn_hold & 0b00000001)						// if the user is holding it, increment duration by 1s
			{
				delay_ms_(BTN_REPEAT_RATE/2);				// slow down, but repeat faster than usual
			}

			btn_press &= ~0b00000001; 					// button handled
			if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
			sound_beep(100,5);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload
		}

		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 					// button handled
			btn_hold &= ~0b00000010;						// also handle the hold
			setup_part++;									// next part
			sound_beep(100,5);								// tick
			sound_beep(20,15);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload to stay in programming mode
		}
	}

	return;
}
// set tubes fading value and on/off for DATE display
void set_date_fading(void)
{
	uint8_t setup_part = 0; 								// first we are going to setup the HH

	btn_inactivity_timer = BTN_INACTIVITY_TMR;				// if the user forgets to click around

	while(setup_part < 2)									// brightness and on/off = 2 things to set in total
	{
		if(btn_inactivity_timer==0) break;					// get out at the exit point of this function

		// brightness?
		if(setup_part==0)
		{
			set_brightness_of(DISPLAY_SIZE,date_fading_brightness);	// set all to this

			// current brightness value: X X X X
			display_digits[0] = (date_fading_brightness / 1000);				// thousands
			display_digits[1] = ((date_fading_brightness % 1000) / 100); 		// hundreds
			display_digits[2] = (((date_fading_brightness % 1000) % 100) / 10); // tens
			display_digits[3] = (((date_fading_brightness % 1000) % 100) % 10); // ones

			rgb_led(0,0,0,255);										// blue all
			rgb_led(1,0,0,255);
			rgb_led(2,0,0,255);
			rgb_led(3,0,0,255);
		}
		// on/off?
		else
		{
			set_brightness_of(DISPLAY_SIZE,max_brightness);	// restore to max configured

			// current option value: x x x X
			display_digits[0] = display_digits[1] = display_digits[2] = nixie_blank;// blank first three digits
			display_digits[3] = (bv(DateFading,BUSR1) ? 1 : 0);	// show current value

			rgb_led(0,0,0,0);										// off with the first three
			rgb_led(1,0,0,0);
			rgb_led(2,0,0,0);
		}

		// BTN-A? held or pressed
		if( (btn_press|btn_hold) & 0b00000001)
		{
			if(btn_hold & 0b00000001)						// if the user is holding it, increment duration by 1s
			{
				// brightness?
				if(setup_part==0)
				{
					// go slower after 4000 :)
					if(date_fading_brightness>=4000)
					{
						delay_ms_(BTN_REPEAT_RATE);			// slow down
						date_fading_brightness+=5;			// increment it!
					}
					// else go very fast
					else
					{
						delay_ms_(BTN_REPEAT_RATE/2);		// slow down, but go faster than usual
						date_fading_brightness+=50;			// increment it!
					}
				}
				// on/off?
				else
				{
					bt(DateFading,BUSR1);					// toggle the value
					delay_ms_(BTN_REPEAT_RATE);				// slow down
				}
			}
			// key pressed
			else
			{
				// brightness?
				if(setup_part==0)
				{
					date_fading_brightness++;				// increment it!
				}
				// on/off?
				else
				{
					bt(DateFading,BUSR1);					// toggle the value
				}
			}

			if(date_fading_brightness>DISP_MAX_BRIGHTNESS) date_fading_brightness=DISP_MIN_BRIGHTNESS; // rollover to min allowed

			btn_press &= ~0b00000001; 					// button handled
			if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
			sound_beep(100,5);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload
		}

		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 					// button handled
			btn_hold &= ~0b00000010;						// also handle the hold
			setup_part++;									// next part
			sound_beep(100,5);								// tick
			sound_beep(20,15);								// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 		// reload to stay in programming mode
		}
	}

	return;
}
// set auto-off times and on/off
void set_auto_off_prefs()
{
	uint8_t setup_part = 0; 								// first we are going to setup the HH

	btn_inactivity_timer = BTN_INACTIVITY_TMR;				// if the user forgets to click around

	// both dots on
	set1(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
	set1(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);

	while(setup_part < 5)									// setting HH, MM, and again HH, MM and on/off
	{
		if(btn_inactivity_timer==0) break;					// get out at the exit point of this function

		if(setup_part == 0 || setup_part == 2)
		{
			rgb_led(0,0,0,255);								// blue us up
			rgb_led(1,0,0,255);
			rgb_led(2,0,0,0);								// off us
			rgb_led(3,0,0,0);
		}
		else if(setup_part == 1 || setup_part == 3)
		{
			rgb_led(0,0,0,0);								// off us
			rgb_led(1,0,0,0);
			rgb_led(2,0,0,255);								// blue us up
			rgb_led(3,0,0,255);
		}

		if(setup_part == 0 || setup_part == 1)
		{
			// show the value: hh:mm
			display_digits[0]=((RTC_AOFF[TIME_H] % 24) / 10);// [H]H:MM
			display_digits[1]=((RTC_AOFF[TIME_H] % 24) % 10);// H[H]:MM
			display_digits[2]=(RTC_AOFF[TIME_M] / 10);		// HH  :[M]M
			display_digits[3]=(RTC_AOFF[TIME_M] % 10);		// HH  :M[M]
		}
		else if(setup_part == 2 || setup_part == 3)
		{
			// show the value: hh:mm
			display_digits[0]=((RTC_AON[TIME_H] % 24) / 10);// [H]H:MM
			display_digits[1]=((RTC_AON[TIME_H] % 24) % 10);// H[H]:MM
			display_digits[2]=(RTC_AON[TIME_M] / 10);		// HH  :[M]M
			display_digits[3]=(RTC_AON[TIME_M] % 10);		// HH  :M[M]
		}
		else
		{
			rgb_led(0,0,0,0);								// off us
			rgb_led(1,0,0,0);
			rgb_led(2,0,0,0);
			rgb_led(3,0,0,255);								// blue this one

			// both dots off
			set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
			set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);

			display_digits[0] = display_digits[1] = display_digits[2] = nixie_blank;
			display_digits[3] = ( bv(AutoOffMode,BUSR1) ? 1:0 );
		}

		// BTN-A? held or pressed
		if( (btn_press|btn_hold) & 0b00000001)
		{
			switch(setup_part)
			{
				// the AOFF HH:mm
				case 0:
					RTC_AOFF[TIME_H] = (RTC_AOFF[TIME_H]+1) % 24; // +1
				break;
				// the AOFF hh:MM
				case 1:
					RTC_AOFF[TIME_M] = (RTC_AOFF[TIME_M]+1) % 60; // +1
				break;
				// the AON HH:mm
				case 2:
					RTC_AON[TIME_H] = (RTC_AON[TIME_H]+1) % 24; // +1
				break;
				// the AON hh:MM
				case 3:
					RTC_AON[TIME_M] = (RTC_AON[TIME_M]+1) % 60; // +1
				break;
				// the on/off part
				case 4:
					bt(AutoOffMode,BUSR1);						// toggle the value
				break;
			}

			if(btn_hold & 0b00000001)							// if the user is holding it, increment hour by 1s
			{
				delay_ms_(BTN_REPEAT_RATE);						// slow down but not too much
			}

			btn_press &= ~0b00000001; 						// button handled
			if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
			sound_beep(100,5);									// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 			// reload
		}

		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 						// button handled
			btn_hold &= ~0b00000010;							// also handle the hold
			setup_part++;
			sound_beep(100,5);									// tick
			sound_beep(20,15);									// tick
			btn_inactivity_timer = BTN_INACTIVITY_TMR; 			// reload to stay in programming mode
		}

	}

	// if auto-off is not used (anymore) cancel automatic auto-off executer, just in case!
	if( !bv(AutoOffMode,BUSR1) )
	{
		auto_off_auto_executer=0;
	}

	return;
}
// set cross-fader speed and on/off
void set_cross_fader_prefs(void)
{
	#ifdef CROSS_FADING_CODE
		uint8_t setup_part = 0; 								// first we are going to setup the HH

		btn_inactivity_timer = BTN_INACTIVITY_TMR;				// if the user forgets to click around

		while(setup_part < 2)									// setting speed and on/off
		{
			if(btn_inactivity_timer==0) break;					// get out at the exit point of this function

			if(setup_part == 0)
			{
				rgb_led(0,0,0,255);								// blue us
				rgb_led(1,0,0,255);
				rgb_led(2,0,0,255);
				rgb_led(3,0,0,0);								// off this one

				// display current value XXXn
				display_digits[0] = (cross_fade_speed / 100); 		// hundreds
				display_digits[1] = ((cross_fade_speed % 100) / 10); // tens
				display_digits[2] = ((cross_fade_speed % 100) % 10); // ones
				display_digits[3] = nixie_blank;
			}
			else
			{
				rgb_led(0,0,0,0);								// off us
				rgb_led(1,0,0,0);
				rgb_led(2,0,0,0);
				rgb_led(3,0,0,255);								// blue this one

				display_digits[0] = display_digits[1] = display_digits[2] = nixie_blank;
				display_digits[3] = ( bv(CrossFading,BUSR1) ? 1:0 );
			}

			// BTN-A? held or pressed
			if( (btn_press|btn_hold) & 0b00000001)
			{
				switch(setup_part)
				{
					// the AOFF HH:mm
					case 0:
						cross_fade_speed = (cross_fade_speed+1) % ((DISP_MAX_BRIGHTNESS-DISP_LAST_OFF_BRTNS)/4); // +1, the maximum value is 1/4*(DISP_MAX_BRIGHTNESS-DISP_LAST_OFF_BRTNS) because that's the range that makes sense in cross-fader
					break;
					// the on/off part
					case 1:
						bt(CrossFading,BUSR1);						// toggle the value
					break;
				}

				if(btn_hold & 0b00000001)							// if the user is holding it, increment hour by 1s
				{
					delay_ms_(BTN_REPEAT_RATE);						// slow down but not too much
				}

				btn_press &= ~0b00000001; 						// button handled
				if( BTNA_PINREG & _BV(BTNA_PIN) ) btn_hold &= ~0b00000001; // if it is held no more, we can handle it
				sound_beep(100,5);									// tick
				btn_inactivity_timer = BTN_INACTIVITY_TMR; 			// reload
			}

			// BTN-B?
			if(btn_press & 0b00000010)
			{
				btn_press &= ~0b00000010; 						// button handled
				btn_hold &= ~0b00000010;							// also handle the hold
				setup_part++;
				sound_beep(100,5);									// tick
				sound_beep(20,15);									// tick
				btn_inactivity_timer = BTN_INACTIVITY_TMR; 			// reload to stay in programming mode
			}

		}
	#endif

	return;
}

// activate user preferences to application
void restore_setup(void)
{
	// restore stuff from user preferences
	if( bv(LEDtoggling,BUSR1) ) bs(ToggleLEDs,BAPP1);
	if( bv(CrossFading,BUSR1) ) bs(CrossFaderEn,BAPP2);

	return;
}
// makes a real color for a given temperature
uint8_t rgb_blender(uint8_t ColorMax, uint8_t ColorMin, int8_t temp)
{
	// boundaries
	if(temp <= (int8_t)Tmin) return ColorMin; 			// return color for minimum temperature
	if(temp >= (int8_t)Tmax) return ColorMax; 			// return color for maximum temperature
	if(ColorMin == ColorMax) return ColorMin; 			// nothing to do here, return any

	// now find a color value between given values ColorMin and ColorMax .. thanks Emina Z. for the formula :)
	float Kt = (float)(ColorBlender_Kb * (Tmax-Tmin)) / (float)(ColorMax-ColorMin);
	float rez = ColorMin + (float)(ColorBlender_Kb/Kt) * (float)(temp-Tmin);

	return (uint8_t)rez; 								// between ColorMax and ColorMin
}
// moving to next thing to display
void disp_next()
{
	if(++disp_what >= DISP_THINGS) disp_what=0;			// next, and check end
	milliseconds = 0;									// for detecting the endless loop
	while(disp_duration[disp_what] == 0)				// in case this one is off, try next one
	{
		if(milliseconds > 50)							// endless loop detected? (user turned off everything by mistake in setup)
		{
			// all were off, enable "time mode" and exit loop
			disp_duration[0] = DEFAULT_DISP_0;
			disp_what = 0;
			break;
		}
		if(++disp_what >= DISP_THINGS) disp_what=0;		// next, and check end
	}
	disp_duration_timer = 0;							// !
}
// chasing LEDs with fade effect with a custom color (NOTE: delay is not very acurate)
void faded_chase_leds(uint8_t r_min, uint8_t g_min, uint8_t b_min, uint8_t r_max, uint8_t g_max, uint8_t b_max, uint8_t steps, uint16_t total_delay_ms, uint8_t dir)
{
	uint8_t fading_steps = (steps+1) * 2;				// calculate how many fading steps we will have for chosen step size
	uint8_t i, r_val, g_val, b_val;
	uint8_t k;

	for(i=1; i<=fading_steps; i++)
	{
		for(k=0; k<4; k++)								// 4 LEDs available
		{
			r_val = faded_chase_leds_calc(i,k,r_min,r_max,steps);// calculate for each color
			g_val = faded_chase_leds_calc(i,k,g_min,g_max,steps);
			b_val = faded_chase_leds_calc(i,k,b_min,b_max,steps);

			uint8_t idx = k;
			// right to left override?
			if(dir==2)
			{
				//idx=6-(3+k);
				idx=3-k; // TODO: check to see if this is OK!
			}

			rgb_led(idx,r_val,g_val,b_val);				// put value at once
		}
		delay_ms_(total_delay_ms/fading_steps);			// make a delay between changing values among LEDs
	}
	for(k=0; k<4; k++)									// 4 LEDs available
	{
		rgb_led(k,0,0,0);								// all off on return
	}
}
// a "helper" function of faded_chase_leds() function
uint8_t faded_chase_leds_calc(uint8_t i, uint8_t k, uint8_t x_min, uint8_t x_max, uint8_t steps)
{
	uint8_t stepval = x_max / steps;					// find values of a step for this color

	if(x_max==0) return 0;								// get out for a zero value
	if(x_max <= x_min) return x_min;					// nothing to do also

	int16_t val = (stepval*i) - (stepval*k);			// do the job
	if(val > x_max) val = (x_max*2) - val;
	if(val < x_min) val = x_min;

	return (uint8_t)val;								// convert to 0-255
}
// special mode for a bit of romance
void candle_mode(void)
{
	uint16_t tmp;

	bc(ToggleColons,BAPP1); 								// don't need this
	bc(ToggleLEDs,BAPP1); 									// this is bad also
	bc(CrossFaderEn,BAPP2);									// and this

	set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);					// off these
	set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);

	// put all ones on display
	display_digits[0]=display_digits[1]=display_digits[2]=display_digits[3]=1;
	//R=100,G=70,B=0); // this is supposed to be some kind of dark orange color :)

	rgb_led(0,0,0,0);										// off LEDs
	rgb_led(1,0,0,0);
	rgb_led(2,0,0,0);
	rgb_led(3,0,0,0);

	uint8_t i=0;
	while(1)
	{
		tmp = (rand() / 10);
		if( !(milliseconds % 140) )							// on every ?ms put new brightness value
		{
			if(i>=DISPLAY_SIZE) i=0;
			brightness[i] = tmp;
			i++;
		}

		// BTN-A?
		if( btn_press & 0b00000001)
		{
			btn_press &= ~0b00000001; 						// button handled
			btn_hold &= ~0b00000001;						// also handle the hold
			break;											// exit at the exit point of this function

		}
		// BTN-B?
		if(btn_press & 0b00000010)
		{
			btn_press &= ~0b00000010; 						// button handled
			btn_hold &= ~0b00000010;						// also handle the hold
			break;											// exit at the exit point of this function
		}
	}

	// activate user preferences to application
	restore_setup();

	set_brightness_of(DISPLAY_SIZE,max_brightness);

	return;
}


//########################
// Pin interrupts setup //
//########################
// init IO pins used with interrupts and set them
void pin_interrupt_init()
{
	// configure used uC pins

	// power sense pin
	PWR_SENS_DDR &= ~_BV(PWR_SENS_PIN); 				// power sense pin input
	PWR_SENS_PORT &= ~_BV(PWR_SENS_PIN); 				// turn OFF internal pullup for this pin, it actually needs pull-down resistor
	PWR_SENS_PCMSKREG |= _BV(PWR_SENS_PCINTBIT); 		// set (un-mask) PCINTn pin for interrupt on change
	PCICR |= _BV(PWR_SENS_PCICRBIT); 					// enable wanted PCICR

	// button A pin
	BTNA_DDR &= ~_BV(BTNA_PIN); 						// button A pin input
	BTNA_PORT |= _BV(BTNA_PIN); 						// turn ON internal pullup
	BTNA_PCMSKREG |= _BV(BTNA_PCINTBIT); 				// set (un-mask) PCINTn pin for interrupt on change
	PCICR |= _BV(BTNA_PCICRBIT); 						// enable wanted PCICR

	// button B pin
	BTNB_DDR &= ~_BV(BTNB_PIN); 						// button B pin input
	BTNB_PORT |= _BV(BTNB_PIN); 						// turn ON internal pullup
	BTNB_PCMSKREG |= _BV(BTNB_PCINTBIT); 				// set (un-mask) PCINTn pin for interrupt on change
	PCICR |= _BV(BTNB_PCICRBIT); 						// enable wanted PCICR

	// alarm switch pin
	SWALARM_DDR &= ~_BV(SWALARM_PIN); 					// alarm switch pin input
	SWALARM_PORT |= _BV(SWALARM_PIN); 					// turn ON internal pullup
	SWALARM_PCMSKREG |= _BV(SWALARM_PCINTBIT); 			// set (un-mask) PCINTn pin for interrupt on change
	PCICR |= _BV(SWALARM_PCICRBIT); 					// enable wanted PCICR

	return;
}


//################################
// Nixie tube related functions //
//################################
// if tube_idx >= DISPLAY_SIZE then brightness is set to all available tubes
void set_brightness_of(uint8_t tube_idx, uint16_t val)
{
	if(tube_idx >= DISPLAY_SIZE)
	{
		uint8_t i;
		for(i=0; i<DISPLAY_SIZE; i++)
		{
			brightness[i] = val;
		}
	}
	else
	{
		brightness[tube_idx] = val;
	}
}


//#############################
// SN74141 related functions //
//#############################
// sets uC pins and HW
void sn74141_init()
{
	// set to output pins we are using
	SN74141_DDR |= SN74141_PINS;

	// clear all on SN's port
	SN74141_PORT &= ~SN74141_PINS;

	// here we also init the two Nixie dot outputs
	// set them to output
	NIXIE_UDOT_DDR |= _BV(NIXIE_UDOT_PIN);
	NIXIE_LDOT_DDR |= _BV(NIXIE_LDOT_PIN);
	// turn them off
	NIXIE_UDOT_PORT |= _BV(NIXIE_UDOT_PIN);
	NIXIE_LDOT_PORT |= _BV(NIXIE_LDOT_PIN);

	return;
}
// sets one of 10 outputs based on numerical input; digit is 0..9
void sn74141_set(uint8_t digit)
{
	if(digit > 10) digit = 10; 								// invalid values (>=10) are blank

	// clear all on SN's port
	SN74141_PORT &= ~SN74141_PINS;

	// now set just ones that need to be set
	SN74141_PORT |= pgm_read_byte(sn74141_map + digit);

	return;
}


//##########################
// Interrupt driven delay //
//##########################
// interrupt based delay function
void delay_ms_(uint16_t ms) {
  milliseconds = 0;
  while (milliseconds < ms);

  return;
}


//#################
// RTC functions //
//#################
// inits the timers for RTC (Timer/Counter2)
void rtc_init()
{
	RTC[TIME_H] = 12;
	RTC[TIME_M] = 0;
	RTC[TIME_S] = 0;

	RTC[DATE_Y] = 10; 										// YYY format.. 10=2010, 124=2124,...
	RTC[DATE_M] = 9;
	RTC[DATE_D] = 1;

	RTC_ALARM[TIME_H] = 0;
	RTC_ALARM[TIME_M] = 0;

	bc(RTC_AlarmOn,BAPP1); 									// alarm is off by default
	bc(RTC_Alarming,BAPP1); 								// and we are not alarming now
	bc(RTC_Snoozing,BAPP1);									// not snoozing also

	// use external LP crystal
	// we will overflow once a second, and call an interrupt
	ASSR |= _BV(AS2); 										// use crystal

	TCCR2B |= _BV(CS22) | _BV(CS20); 						// prescale by 128, for 32.768 kHz

	// enable interrupt
	TIMSK2 |= _BV(TOIE2);

	return;
}
// calculate if given year is leap year
uint8_t isleapyear(uint16_t y)
{
  return ( ( !(y % 4) && (y % 100) ) || !(y % 400) );
}
// turning alarm on
void alarm_on()
{
	snooze_auto_executer = 0;
	snooze_timer = 0;

	bs(RTC_AlarmOn,BAPP1);
	bc(RTC_Alarming,BAPP1);
	bc(RTC_Snoozing,BAPP1);

	// don't notify anything if in sleep mode
	if( bv(SleepMode,BSYS) )
		return;

	bc(ToggleLEDs,BAPP1);									// this is bad, so off it
	bc(CrossFaderEn,BAPP2);									// and this

	display_digits[3] = (RTC_ALARM[TIME_M] % 10); 			// HH  :M[M]
	display_digits[2] = (RTC_ALARM[TIME_M] / 10); 			// HH  :[M]M
	display_digits[1] = ((RTC_ALARM[TIME_H] % 24) % 10); 	// H[H]:MM
	display_digits[0] = ((RTC_ALARM[TIME_H] % 24) / 10); 	// [H]H:MM

	sound_beep(150,25);										// alert even with sound
	sound_beep(70,30);
	sound_beep(20,40);
	sound_beep(0,40);
	sound_beep(20,40);

	faded_chase_leds(0,0,0, 255,0,0, 25, 2500, 1);			// red color fading :)

	// activate user preferences to application
	restore_setup();

	return;
}
// turning alarm off
void alarm_off()
{
	snooze_auto_executer = 0;
	snooze_timer = 0;

	bc(RTC_AlarmOn,BAPP1);
	bc(RTC_Alarming,BAPP1);
	bc(RTC_Snoozing,BAPP1);

	// don't notify anything if in sleep mode
	if( bv(SleepMode,BSYS) )
		return;

	bc(ToggleLEDs,BAPP1);									// this is bad, so off it
	bc(CrossFaderEn,BAPP2);									// and this

	sound_beep(20,25);										// alert even with sound

	sound_beep(70,30);
	sound_beep(150,40);
	sound_beep(0,40);
	sound_beep(150,40);

	faded_chase_leds(0,0,0, 0,20,255, 15, 2500, 1);			// cool blue color fading :)

	// activate user preferences to application
	restore_setup();

	return;
}
// snoozing the alarm
void alarm_snooze()
{
	snooze_timer = snooze_interval * 60; 					// to seconds
	bs(RTC_Snoozing,BAPP1);									// say that we are snoozing and when "snooze_timer" gets to 0, sound the alarm again
	bc(RTC_Alarming,BAPP1);									// this will tell ISR to turn off the alarm... for now :)

	// report snoozing here
	bc(ToggleLEDs,BAPP1);									// this is bad, so off it
	bc(CrossFaderEn,BAPP2);									// and this

	faded_chase_leds(0,0,0, 20,80,255, 10, 2000, 1);		// snooze color fading :)

	// activate user preferences to application
	restore_setup();

	return;
}
// set melody to play
void alarm_play(void)
{
	// added 02-11-2010 - if we are asleep, this will wake us up
	boost_continue();
	//--

	snooze_timer = 0;
	snooze_auto_executer=0;
	bs(RTC_Alarming,BAPP1);
	sound_duration = 1; 									// just to enter the ISR's loop
	bc(AutoOffed,BSYS);										// wake the clock up until next sleep time-match!

	return;
}

//##############################
// TLC59401 related functions //
//##############################
// sets uC pins and HW
void tlc59401_init() {
	// ddr - outputs
	GSCLK_DDR |= _BV(GSCLK_PIN);

	SCLK_DDR |= _BV(SCLK_PIN);
	SIN_DDR |= _BV(SIN_PIN);

	MODE_DDR |= _BV(MODE_PIN);
	XLAT_DDR |= _BV(XLAT_PIN);
	BLANK_DDR |= _BV(BLANK_PIN);

	// pin-states
	set0(GSCLK_PORT, GSCLK_PIN);

	set0(SCLK_PORT, SCLK_PIN);
	set0(SIN_PORT, SIN_PIN);

	set1(MODE_PORT, MODE_PIN);
	set0(XLAT_PORT, XLAT_PIN);
	set1(BLANK_PORT, BLANK_PIN);

	// set the interrupt Timer1 counter A to execute on 4096 cycles
	TCCR1A = 0;
	TCCR1B = _BV(WGM12) | _BV(CS10); 					// CTC (Clear Timer on Compare match), top is at OCR1A, reload Immediate, set flag on MAX | no prescaling, timer started

	TIMSK1 = _BV(OCIE1A); 								// enable interrupt for OCR1A compare match value

	OCR1A = 4095; 										// on every 4096 cycles we must reset the BLANK pin on TLC

	//EnableSPI,Master,setclockratefck/2
	SPDR=0;
	SPCR= _BV(SPE) | _BV(MSTR);
	SPSR= _BV(SPI2X);

	return;
}
// sending DC data to TLC chip
void tcl59401_ClockInDC()
{
	while( bv(TLC_HasDCUpdate,BSYS) || bv(TLC_HasGSUpdate,BSYS) ); // here we wait for TLC to be ready to accept new data and if something is waiting to be latched in TLC
	set1(MODE_PORT,MODE_PIN); 							// set to DC

	uint8_t i;
	for(i=0;i<12;i++)
	{
		SPDR=tlc_DCdata[i]; 							// start the transmission
		while(!(SPSR&(1<<SPIF))); 						// wait for the transmission to complete
	}

	bs(TLC_NeedsExtraCLK,BSYS); 						// tell interrupt that after next GS data it needs to issue additional CLK impulse (as per datasheet)
	bs(TLC_HasDCUpdate,BSYS); 							// tell interrupt we sent DC data to the TLC

	return;
}
// sending GS data to TLC chip
void tcl59401_ClockInGS()
{
	while( bv(TLC_HasDCUpdate,BSYS) || bv(TLC_HasGSUpdate,BSYS) ); // here we wait for TLC to be ready to accept new data and if something is waiting to be latched in TLC
	set0(MODE_PORT, MODE_PIN); 							// set to GS

	uint8_t i;
	for(i=0;i<24;i++)
	{
		SPDR=tlc_GSdata[i]; 							// start the transmission
		while(!(SPSR&(1<<SPIF))); 						// wait for the transmission to complete
	}

	bs(TLC_HasGSUpdate,BSYS); 							// tell interrupt we sent GS data to the TLC

	return;
}
// sets GS data for a channel in data array
void tlc59401_SetGSdata(uint8_t ch, uint16_t val)
{
	ch = 15 - (uint16_t)ch;
	uint8_t idx = (uint8_t)((uint16_t)ch * 3 / 2);				// calculates channel's index in our data array, index moves per 1.5 bytes
	val &= 0x0FFF;												// to make sure this doesn't overflow above 4095
	uint8_t *arr = (uint8_t *)tlc_GSdata + idx;
	// if the channel is odd, meaning the value starts in the middle of the byte in our tlc_GSdata[] data array
	if(ch & 0b00000001)
	{
		*arr = (*arr & 0b11110000) | (uint8_t)(val >> 8);		// copy just 4 bits now to lower nibble of destination data array
		arr++;
		*arr= (uint8_t)val & 0b11111111;						// now copy remaining 8 bits to next location .. 12 in total
	}
	// nope, it is an even channel
	else
	{
		*(arr++) = (uint8_t)(val >> 4);							// copy entire 8 bits to this location of destination array
		*arr = (uint8_t)(val << 4) | (*arr & 0b00001111);		// now copy remaining 4 bits to next location .. 12 in total
	}

	return;
}
// sets "nr" RGB LED to 16* r,g,b value; nr is 0..3
void rgb_led(uint8_t nr, uint8_t r, uint8_t g, uint8_t b)
{
	// we have 4 RGB LEDs available here, scattered on 12 channels, let's find them in PROGMEM table
	tlc59401_SetGSdata(pgm_read_byte(rgb_tlc_map + nr + 0), 16*r); //red (bug: 0x0FF0 max value, it should be 0x0FFF max)
	tlc59401_SetGSdata(pgm_read_byte(rgb_tlc_map + nr + 4), 16*g); //green (bug: 0x0FF0 max value, it should be 0x0FFF max)
	tlc59401_SetGSdata(pgm_read_byte(rgb_tlc_map + nr + 8), 16*b); //blue (bug: 0x0FF0 max value, it should be 0x0FFF max)

	return;
}


//########################
// HV booster functions //
//########################
// sets uC pins and HW
void boost_init()
{
	// configure used uC pins
	BOOST_DDR |= _BV(BOOST_PIN); 						// boost pin output
	BOOST_PORT &= ~_BV(BOOST_PIN); 						// pull boost fet low so we don't make short-circuit

	// DDR feedback A/D input pin
	FEEDBACK_DDR &= ~_BV(FEEDBACK_PIN);

	// configure PWM output
	TCCR0A |= _BV(COM0B1); 								// Clear OC0B on Compare Match, set OC0B at BOTTOM, (non-inverting mode)
	TCCR0A |= _BV(WGM01) | _BV(WGM00); 					// Fast PWM mode 3
	TCCR0B  = _BV(CS00); 								// no prescaling, timer started

	TIMSK0 |= _BV(TOIE0); 								//-for-> SIGNAL(SIG_OVERFLOW0)

	bc(BoosterOnline,BSYS);

	return;
}
// starts the booster, returns 0 on success, 1 on failure, 2 on excess voltage
uint8_t boost_go(uint8_t wanted_voltage, uint8_t maximum_voltage)
{
	#ifdef DEBUG_BOOST_VAL
		OCR0B = DEBUG_BOOST_VAL;
	#else
		uint8_t feedback_voltage = 0;

		OCR0B = 5; 										// start at 5 * 0,000000125 sec

		// loop untill success or die
		while(feedback_voltage < wanted_voltage)
		{
			// if i can't generate wanted_voltage, return with error
			if(OCR0B > 245) 							// dont go after 245=96% because the coil will not have enough time to discharge
			{
				boost_halt(); 							// stop booster hardware because it didn't work out for us
				return 1; 								// return from function with error
			}

			OCR0B++; // try with more? ++0,000000125 sec
			delay_ms_(10);

			feedback_voltage=boost_feedback(15); 		// get current generated voltage value
		}
		// check for too much voltage and decrement OCR0B by 1
		if(feedback_voltage > maximum_voltage)
		{
			OCR0B--; 									// hopefully this helped
			return 2; 									// say that there was overvoltage rather than wanted voltage
		}
	#endif

	return 0;
}
// stops the booster
void boost_halt()
{
	OCR0B = 0;

	TCCR0A &= ~_BV(WGM01) & ~_BV(WGM00); 				// Timer0 still works, but ORC0A is disconnected: Mode 0
	BOOST_PORT &= ~_BV(BOOST_PIN); 						// pull boost fet low so we don't make short-circuit, just in case PWM didn't

	bc(BoosterOnline,BSYS);								// say that the booster is not online

	return;
}
// re-enables the booster PWM
void boost_continue()
{
	if( bv(BoosterOnline,BSYS) ) return;				// if booster is running then exit

	// configure PWM output
	TCCR0A |= _BV(COM0B1); 								// Clear OC0B on Compare Match, set OC0B at BOTTOM, (non-inverting mode)
	TCCR0A |= _BV(WGM01) | _BV(WGM00); 					// Fast PWM mode 3

	boost_restart();									// generate voltage...

	return;
}
// gets booster output voltage on call
uint8_t boost_feedback(uint8_t how_many)
{
	uint8_t i;
	uint16_t volt_sum = 0;
	uint16_t current_voltage = 0;

	if(how_many>63) how_many=63; 						// (65535/1023=64) we can't store more than this many values
	if(how_many==0) how_many++; 						// because of the division with it

	// set ADMUX channel and reference voltage
	ADMUX = FEEDBACK_ADMUX_VAL; 						// AREF (Internal Vref turned off), ADC1

	// initialize the ADC circuit
	ADCSRA = _BV(ADEN) | _BV(ADPS2) | _BV(ADPS1); 		// enabled, division factor: 64

	// no effect since ADATE in ADCSRA is = 0
	ADCSRB = 0x00;

	// disable digital input on ADC1
	FEEDBACK_DIDR |= FEEDBACK_DIDR_VAL;

	// make "n" measurements, find average value
	for(i=0; i<how_many; i++)
	{
		// Start conversion by setting ADSC on ADCSRA Register
		ADCSRA |= _BV(ADSC);

		// wait until convertion complete ADSC=0 -> Complete
		while (ADCSRA & (1<<ADSC));

		// Sum the current voltage
		current_voltage = ADCW;
		volt_sum += current_voltage;
	}
	volt_sum /= how_many;

	// convert to volts and return
	// we know that Vref is at ~4.82
	// formula: volt_sum *(4+(86/100)) /1024) *(58+(93/100); // 58.93=330k/5.6k (voltage divider)
	//return (uint8_t)(volt_sum * 5 / 1024) * 59; 	// works but not accurate at all !!!
	//return (uint8_t)(volt_sum *4.8 /1024.0 *(330.0/5.6)); // accurate but FLASH hungry !!!
	return (uint8_t)(volt_sum *4.76 /1024.0 *(820.0/10.0)); // 01-12-2010 changed 330k->820k and 5k6->10k resistors in FEEDBACK voltage divider... accurate but FLASH hungry !!!
}
// this encapsulates booster functions and also reports result on LEDs and Nixies
void boost_restart()
{
	if( bv(BoosterOnline,BSYS) ) return;				// if booster is running then exit

	// put the most current consuming digits on on all nixies for voltage booster adjustment
	display_digits[0] = display_digits[2] = 9;
	display_digits[1] = display_digits[3] = 8;
	set1(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN); 				// dot on
	set1(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN); 				// dot on

	uint8_t temp; 										// booster's return value info/error, and for voltage reporting on Nixies
	if( (temp=boost_go(185,190)) != 0) 					// start it. note: before starting the booster, we must enable the TLC chip so that Nixies are ON
	{
		uint8_t i, j;
		// report booster info/error here
		switch(temp)
		{
			case 1:
				// there was an error with booster, wanted voltage couldn't be produced. blink RED on all LEDs and beep for some time
				for(j=0; j<10; j++)
				{
					sound_beep(150,25);
					sound_beep(50,25);
					faded_chase_leds(0,0,0, 250,0,0, 10, 1000, 2);
				}
			break;
			case 2:
				// there was overvoltage rather than wanted voltage, so it is reduced to next best value. orange all LEDs for little time
				for(i=0; i<3; i++)
				{
					sound_beep(50,25);
					sound_beep(150,25);
					faded_chase_leds(0,0,0, 200,50,0, 10, 1000, 2);
				}
			break;
		}
	}

	// If we don't use the "smart" booster, don't report the voltage back on nixies
	#ifndef DEBUG_BOOST_VAL
		// Show the booster voltage on Nixies
		temp = boost_feedback(10); 						// get the voltage
		set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN); 			// dot off
		set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN); 			// dot off
		display_digits[0] = nixie_blank; 				// blank first digit
		display_digits[1] = (temp / 100); 				// hundreds
		display_digits[2] = ((temp % 100) / 10); 		// tens
		display_digits[3] = ((temp % 100) % 10); 		// ones
		delay_ms_(3000); 								// show it for 3 seconds
	#endif

	bs(BoosterOnline,BSYS);								// say that the booster is running

	return;
}

//##############################
// Sound generation functions //
//##############################
// sets uC pins and HW
void sound_init()
{
	// configure used uC pins
	SOUND_DDR |= _BV(SOUND_PIN);
	SOUND_PORT &= ~_BV(SOUND_PIN);

	sound_duration = 0; 								// no sound to produce now

	return;
}
// beeps @freq for dur duration (ms)
void sound_beep(uint8_t freq, uint16_t dur)
{
	while(sound_duration); 								// wait until it finishes playing current note

	sound_freq = freq;									// tell the ISR
	sound_duration = dur;								// this baby will start the sound generator!

	return;
}


//#############################
// Sleep start/end functions //
//#############################
// sets all to sleep mode, turn off some peripherals
void sleep_go()
{
	// leave only pin-change interrupt enabled, just the PWR_SENS one

	boost_halt();

	// clear all outputs that external chips use !
	SN74141_PORT &= ~SN74141_PINS;

	set0(NIXIE_UDOT_PORT,NIXIE_UDOT_PIN);
	set0(NIXIE_LDOT_PORT,NIXIE_LDOT_PIN);

	set0(GSCLK_PORT, GSCLK_PIN);
	set0(SCLK_PORT, SCLK_PIN);
	set0(SIN_PORT, SIN_PIN);
	set0(MODE_PORT, MODE_PIN);
	set0(XLAT_PORT, XLAT_PIN);
	set0(BLANK_PORT, BLANK_PIN);

	bc(RTC_Alarming,BAPP1);							// clear this
	bc(RTC_Snoozing,BAPP1);							// and this

	bs(SleepMode,BSYS);

	BTNA_PCMSKREG &= ~_BV(BTNA_PCINTBIT);			// button A disabled
	BTNB_PCMSKREG &= ~_BV(BTNB_PCINTBIT);			// button B disabled
	SWALARM_PCMSKREG &= ~_BV(SWALARM_PCINTBIT);		// alarm switch disabled

	return;
}
// restores all from sleep mode, turn back on some peripherals
void sleep_end()
{
	// turn back on stuff

	boost_continue();								// restart this baby

	BTNA_PCMSKREG |= _BV(BTNA_PCINTBIT);			// button A re-enabled
	BTNB_PCMSKREG |= _BV(BTNB_PCINTBIT);			// button B re-enabled
	SWALARM_PCMSKREG |= _BV(SWALARM_PCINTBIT);		// alarm switch re-enabled

	bc(SleepMode,BSYS);
	bc(AutoOffed,BSYS);								// in case we turned the power off while the device was auto-offed :)
	auto_off_auto_executer = 0;

	return;
}
